# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import logging
import numpy as np
import math
import traceback

from .util import LoggerWrapper, GetRNG, UpdateNProc, MultiProcess, SetupConfigRNG, RemoveCurrent
from .input import SetupInput
from .gsobject import UpdateGSParams, SkipThisObject
from .gsobject import BuildGSObject
from .value import ParseValue, CheckAllParams
from .noise import CalculateNoiseVariance, AddSky, AddNoise
from .wcs import BuildWCS
from .photon_ops import BuildPhotonOps
from ..errors import GalSimConfigError, GalSimConfigValueError
from ..image import Image
from ..wcs import PixelScale
from ..position import PositionD, PositionI
from ..bounds import _BoundsI
from ..celestial import CelestialCoord
from ..angle import arcsec
from ..gsobject import GSObject
from ..convolve import Convolve, Convolution
from ..box import Pixel
from ..chromatic import ChromaticObject
from ..bandpass import Bandpass

# This file handles the building of postage stamps to place onto a larger image.
# There is only one type of stamp currently, called Basic, which builds a galaxy from
# config['gal'] and a PSF from config['psf'] (either but not both of which may be absent),
# colvolves them together, and draws onto a postage stamp.  This is the default functionality
# and is typically not specified explicitly.  But there are hooks in place to allow for other
# options, either in future versions of GalSim or through user modules.

# This module-level dict will store all the registered stamp types.
# See the RegisterStampType function at the end of this file.
# The keys are the (string) names of the output types, and the values will be builder objects
# that will perform the different stages of processing to build each stamp image.
valid_stamp_types = {}


def BuildStamps(nobjects, config, obj_num=0,
                xsize=0, ysize=0, do_noise=True, logger=None):
    """
    Build a number of postage stamp images as specified by the config dict.

    Parameters:
        nobjects:       How many postage stamps to build.
        config:         A configuration dict.
        obj_num:        If given, the current obj_num. [default: 0]
        xsize:          The size of a single stamp in the x direction. [default: 0,
                        which means to look first for config.stamp.xsize, then for
                        config.image.stamp_xsize, and if neither are given, then use
                        automatic sizing.]
        ysize:          The size of a single stamp in the y direction. [default: 0,
                        which means to look first for config.stamp.ysize, then for
                        config.image.stamp_ysize, and if neither are given, then use
                        automatic sizing.]
        do_noise:       Whether to add noise to the image (according to config['noise']).
                        [default: True]
        logger:         If given, a logger object to log progress. [default: None]

    Returns:
        the tuple (images, current_vars).  Both are themselves tuples.
    """
    logger = LoggerWrapper(logger)
    logger.debug('image %d: BuildStamps nobjects = %d: obj = %d',
                 config.get('image_num',0),nobjects,obj_num)

    if nobjects == 0:
        logger.warning("No stamps were built, since nstamps == 0.")
        return (), ()

    # Figure out how many processes we will use for building the stamps:
    if nobjects > 1 and 'image' in config and 'nproc' in config['image']:
        nproc = ParseValue(config['image'], 'nproc', config, int)[0]
        # Update this in case the config value is -1
        nproc = UpdateNProc(nproc, nobjects, config, logger)
    else:
        nproc = 1

    if 'image' in config and 'timeout' in config['image']:
        timeout = ParseValue(config['image'], 'timeout', config, float)[0]
    else:
        timeout = 900

    jobs = []
    for k in range(nobjects):
        kwargs = {
            'obj_num' : obj_num + k,
            'xsize' : xsize,
            'ysize' : ysize,
            'do_noise' : do_noise,
        }
        jobs.append(kwargs)

    def done_func(logger, proc, k, result, t):
        if result[0] is not None:
            # Note: numpy shape is y,x
            image = result[0]
            ys, xs = image.array.shape
            if proc is None: s0 = ''
            else: s0 = '%s: '%proc
            obj_num = jobs[k]['obj_num']
            logger.info(s0 + 'Stamp %d: size = %d x %d, time = %f sec', obj_num, xs, ys, t)

    def except_func(logger, proc, k, e, tr):
        if proc is None: s0 = ''
        else: s0 = '%s: '%proc
        obj_num = jobs[k]['obj_num']
        logger.error(s0 + 'Exception caught when building stamp %d', obj_num)
        logger.debug('%s',tr)
        logger.error('Aborting the rest of this image')

    # Convert to the tasks structure we need for MultiProcess.
    # Each task is a list of (job, k) tuples.
    tasks = MakeStampTasks(config, jobs, logger)

    results = MultiProcess(nproc, config, BuildStamp, tasks, 'stamp',
                           logger=logger, timeout=timeout,
                           done_func=done_func, except_func=except_func)

    images, current_vars = zip(*results)

    logger.debug('image %d: Done making stamps',config.get('image_num',0))
    if all(im is None for im in images):
        logger.warning('No stamps were built.  All objects were skipped.')

    return images, current_vars

# A list of keys that really belong in stamp, but are allowed in image both for convenience
# and backwards-compatibility reasons.  Any of these present will be copied over to
# config['stamp'] if they exist in config['image'].
stamp_image_keys = ['offset', 'retry_failures', 'gsparams', 'draw_method', 'dtype',
                    'n_photons', 'max_extra_noise', 'poisson_flux', 'obj_rng']

def SetupConfigObjNum(config, obj_num, logger=None):
    """Do the basic setup of the config dict at the stamp (or object) processing level.

    Includes:

    - Set config['obj_num'] = obj_num
    - Set config['index_key'] = 'obj_num'
    - Make sure config['stamp'] exists
    - Set default config['stamp']['type'] to 'Basic'
    - Copy over values from config['image'] that are allowed there, but really belong
      in config['stamp'].
    - Set config['stamp']['draw_method'] to 'auto' if not given.

    Parameters:
        config:         A configuration dict.
        obj_num:        The current obj_num.
        logger:         If given, a logger object to log progress. [default: None]
    """
    logger = LoggerWrapper(logger)
    config['obj_num'] = obj_num
    config['index_key'] = 'obj_num'

    # Make config['stamp'] exist if it doesn't yet.
    if 'stamp' not in config:
        config['stamp'] = {}
    stamp = config['stamp']
    if '_done' in stamp: return

    # Everything after here only needs to be done once.

    if not isinstance(stamp, dict):
        raise GalSimConfigError("config.stamp is not a dict.")
    if 'type' not in stamp:
        stamp['type'] = 'Basic'

    if 'file_num' not in config:
        config['file_num'] = 0
    if 'image_num' not in config:
        config['image_num'] = 0

    # Copy over some things from config['image'] if they are given there.
    # These are things that we used to advertise as being in the image field, but now that
    # we have a stamp field, they really make more sense here.  But for backwards compatibility,
    # or just because they can make sense in either place, we allow them to be in 'image' still.
    if 'image' in config:
        image = config['image']
        for key in stamp_image_keys:
            if key in image and key not in stamp:
                stamp[key] = image[key]
    else:
        config['image'] = {}

    if 'draw_method' not in stamp:
        stamp['draw_method'] = 'auto'

    # In case this hasn't been done yet.
    SetupInput(config, logger)

    # Mark this as done, so later passes can skip a lot of this.
    stamp['_done'] = True

def SetupConfigStampSize(config, xsize, ysize, image_pos, world_pos, logger=None):
    """Do further setup of the config dict at the stamp (or object) processing level reflecting
    the stamp size and position in either image or world coordinates.

    Note: This is now a StampBuilder method.  So this function just calls StampBuilder.locateStamp.

    Parameters:
        config:         A configuration dict.
        xsize:          The size of the stamp in the x-dimension. [may be 0 if unknown]
        ysize:          The size of the stamp in the y-dimension. [may be 0 if unknown]
        image_pos:      The position of the stamp in image coordinates. [may be None]
        world_pos:      The position of the stamp in world coordinates. [may be None]
        logger:         If given, a logger object to log progress. [default: None]
    """
    stamp = config.get('stamp',{})
    stamp_type = stamp.get('type','Basic')
    builder = valid_stamp_types[stamp_type]
    builder.locateStamp(stamp, config, xsize, ysize, image_pos, world_pos, logger)


# Ignore these when parsing the parameters for specific stamp types:
stamp_ignore = ['xsize', 'ysize', 'size', 'image_pos', 'world_pos', 'sky_pos',
                'offset', 'retry_failures', 'gsparams', 'draw_method', 'dtype',
                'n_photons', 'max_extra_noise', 'poisson_flux', 'photon_ops',
                'skip', 'reject', 'min_flux_frac', 'min_snr', 'max_snr',
                'quick_skip', 'obj_rng', 'index_key', 'rng_index_key', 'rng_num']

valid_draw_methods = ('auto', 'fft', 'phot', 'real_space', 'no_pixel', 'sb')

def BuildStamp(config, obj_num=0, xsize=0, ysize=0, do_noise=True, logger=None):
    """
    Build a single stamp image using the given config file

    Parameters:
        config:         A configuration dict.
        obj_num:        If given, the current obj_num [default: 0]
        xsize:          The xsize of the stamp to build (if known). [default: 0]
        ysize:          The ysize of the stamp to build (if known). [default: 0]
        do_noise:       Whether to add noise to the image (according to config['noise']).
                        [default: True]
        logger:         If given, a logger object to log progress. [default: None]

    Returns:
        the tuple (image, current_var)
    """
    from .extra import ProcessExtraOutputsForStamp

    logger = LoggerWrapper(logger)
    SetupConfigObjNum(config, obj_num, logger)

    stamp = config['stamp']
    stamp_type = stamp['type']
    if stamp_type not in valid_stamp_types:
        raise GalSimConfigValueError("Invalid stamp.type.", stamp_type,
                                     list(valid_stamp_types.keys()))
    builder = valid_stamp_types[stamp_type]

    if builder.quickSkip(stamp, config):
        ProcessExtraOutputsForStamp(config, True, logger)
        return None, 0

    builder.setupRNG(stamp, config, logger)

    if 'skip_failures' in stamp:
        skip_failures = ParseValue(stamp, 'skip_failures', config, bool)[0]
    else:
        skip_failures = False

    if 'retry_failures' in stamp:
        ntries = ParseValue(stamp,'retry_failures',config,int)[0]
        if skip_failures and ntries:  # pragma: no cover
            # This is definitely covered by the tests.
            # I can't figure out why codecov doesn't think so.
            raise GalSimConfigValueError(
                "Cannot use retry_failures when skip_failures=True", ntries)
        # This is how many _re_-tries.  Do at least 1, so ntries is 1 more than this.
        ntries = ntries + 1
    elif ('reject' in stamp or 'min_flux_frac' in stamp or
          'min_snr' in stamp or 'max_snr' in stamp):
        # Still impose a maximum number of tries to prevent infinite loops.
        ntries = 20
    else:
        ntries = 1

    itry = 0
    while True:
        itry += 1  # itry increases from 1..ntries at which point we just reraise the exception.

        # The rest of the stamp generation stage is wrapped in a try/except block.
        # If we catch an exception, we continue the for loop to try again.
        # On the last time through, we reraise any exception caught.
        # If no exception is thrown, we simply break the loop and return.
        try:

            # Do the necessary initial setup for this stamp type.
            xsize, ysize, image_pos, world_pos = builder.setup(
                    stamp, config, xsize, ysize, stamp_ignore, logger)

            # Determine the stamp size and location
            builder.locateStamp(stamp, config, xsize, ysize, image_pos, world_pos, logger)

            # Get the global gsparams kwargs.  Individual objects can add to this.
            gsparams = {}
            if 'gsparams' in stamp:
                gsparams = UpdateGSParams(gsparams, stamp['gsparams'], config)

            # Note: Skip is different from Reject.
            #       Skip means we return None for this stamp image and continue on.
            #       Reject means we retry this object using the same obj_num.
            #       This has implications for the total number of objects as well as
            #       things like ring tests that rely on objects being made in pairs.
            #
            #       Skip is also different from prof = None.
            #       If prof is None, then the user indicated that no object should be
            #       drawn on this stamp, but that a noise image is still desired.
            if builder.getSkip(stamp, config, logger):
                raise SkipThisObject('')

            # Build the object to draw
            psf = builder.buildPSF(stamp, config, gsparams, logger)
            prof = builder.buildProfile(stamp, config, psf, gsparams, logger)

            # Make an empty image
            im = builder.makeStamp(stamp, config, xsize, ysize, logger)

            # Determine which draw method to use
            method = builder.getDrawMethod(stamp, config, logger)

            # Determine the net offset (starting from config['stamp_offset'] typically.
            offset = builder.getOffset(stamp, config, logger)

            # At this point, we may want to update whether the object should be skipped
            # based on other information about the profile and location.
            if builder.updateSkip(prof, im, method, offset, stamp, config, logger):
                raise SkipThisObject('')

            # Draw the object on the postage stamp
            im = builder.draw(prof, im, method, offset, stamp, config, logger)
            # Store the final version of the current profile for reference.
            config['current_prof'] = prof

            # Update the drawn image according to the SNR if desired.
            scale_factor = builder.getSNRScale(im, stamp, config, logger)
            im, prof = builder.applySNRScale(im, prof, scale_factor, method, logger)

            # Set the origin appropriately
            builder.updateOrigin(stamp, config, im)

            # Store the current stamp in the base-level config for reference
            config['current_stamp'] = im
            # This is also information that the weight image calculation needs
            config['do_noise_in_stamps'] = do_noise

            # Check if this object should be rejected.
            reject = builder.reject(stamp, config, prof, psf, im, logger)
            if reject:
                if itry < ntries:
                    logger.warning('Object %d: Rejecting this object and rebuilding', obj_num)
                    builder.reset(config, logger)
                    continue
                else:
                    raise GalSimConfigError(
                            "Rejected an object %d times. If this is expected, "
                            "you should specify a larger stamp.retry_failures."%(ntries))

            ProcessExtraOutputsForStamp(config, False, logger)

            # We always need to do the whiten step here in the stamp processing
            current_var = builder.whiten(prof, im, stamp, config, logger)
            if current_var != 0.:
                logger.debug('obj %d: whitening noise brought current var to %f',
                                config.get('obj_num',0),current_var)

            # Sometimes, depending on the image type, we go on to do the rest of the noise as well.
            if do_noise:
                im, current_var = builder.addNoise(stamp,config,im,current_var,logger)

        except SkipThisObject as e:
            if e.msg != '':
                logger.debug('obj %d: Caught SkipThisObject: e = %s',obj_num,e.msg)
            logger.info('Skipping object %d %s', obj_num, e.msg)
            # If xsize, ysize != 0, then this makes a blank stamp for this object.
            # Otherwise, it's just None here.
            im = builder.makeStamp(stamp, config, xsize, ysize, logger)
            ProcessExtraOutputsForStamp(config, True, logger)
            return im, 0.

        except Exception as e:
            if skip_failures:
                logger.info('Object %d: Caught exception %s',obj_num,str(e))
                tr = traceback.format_exc()
                logger.debug('obj %d: Traceback = %s',obj_num,tr)
                logger.info('Skipping this object')
                # Now same as SkipThisObject case above.
                im = builder.makeStamp(stamp, config, xsize, ysize, logger)
                ProcessExtraOutputsForStamp(config, True, logger)
                return im, 0.
            elif itry >= ntries:
                # Then this was the last try.  Just re-raise the exception.
                logger.info('Object %d: Caught exception %s',obj_num,str(e))
                if ntries > 1:
                    logger.error(
                        'Object %d: Too many exceptions/rejections for this object. Aborting.',
                        obj_num)
                raise
            else:
                logger.info('Object %d: Caught exception %s',obj_num,str(e))
                logger.info('This is try %d/%d, so trying again.',itry,ntries)
                tr = traceback.format_exc()
                logger.debug('obj %d: Traceback = %s',obj_num,tr)
                # Need to remove the "current"s from the config dict.  Otherwise,
                # the value generators will do a quick return with the cached value.
                builder.reset(config, logger)
                continue

        else:
            # No exception.
            return im, current_var


def MakeStampTasks(config, jobs, logger):
    """Turn a list of jobs into a list of tasks.

    See the doc string for galsim.config.MultiProcess for the meaning of this distinction.

    For the Basic stamp type, there is just one job per task, so the tasks list is just::

        tasks = [ [ (job, k) ] for k, job in enumerate(jobs) ]

    But other stamp types may need groups of jobs to be done sequentially by the same process.
    cf. stamp type=Ring.

    Parameters:
        config:         The configuration dict
        jobs:           A list of jobs to split up into tasks.  Each job in the list is a
                        dict of parameters that includes 'obj_num'.
        logger:         A logger object to log progress.

    Returns:
        a list of tasks
    """
    stamp = config.get('stamp', {})
    stamp_type = stamp.get('type', 'Basic')
    return valid_stamp_types[stamp_type].makeTasks(stamp, config, jobs, logger)


def DrawBasic(prof, image, method, offset, config, base, logger, **kwargs):
    """The basic implementation of the draw command

    This function is provided as a free function, rather than just the base class implementation
    in StampBuilder to make it easier for classes derived from StampBuilder to use to help
    implement their draw functions.  The base class, StampBuilder, just calls this function
    for its draw method.

    This version also allows for additional kwargs, which are passed on to the drawImage function.
    e.g. you can add add_to_image=True or setup_only=True if these are helpful.

    Parameters:
        prof:       The profile to draw.
        image:      The image onto which to draw the profile (which may be None).
        method:     The method to use in drawImage.
        offset:     The offset to apply when drawing.
        config:     The configuration dict for the stamp field.
        base:       The base configuration dict.
        logger:     A logger object to log progress.
        **kwargs:   Any additional kwargs are passed along to the drawImage function.

    Returns:
        the resulting image
    """
    logger = LoggerWrapper(logger)
    # Setup the kwargs to pass to drawImage
    # (Start with any additional kwargs given as extra kwargs to DrawBasic and add to it.)
    kwargs['image'] = image
    kwargs['offset'] = offset
    kwargs['method'] = method
    if 'wcs' not in kwargs and 'scale' not in kwargs:
        kwargs['wcs'] = base['wcs'].local(image_pos = base['image_pos'])
    sensor = base.get('sensor',None)
    if (method == 'phot' or sensor is not None) and 'rng' not in kwargs:
        # Note: use the image.noise rng_num, in case stamp is doing a non-standard cadence.
        # The cadence specified in the noise field is what to use for photon shooting.
        noise = base.get('image',{}).get('noise',{})
        noise = noise if isinstance(noise, dict) else {}
        rng = GetRNG(noise, base, logger, "method='phot'")
        kwargs['rng'] = rng

    # Check validity of extra phot options:
    max_extra_noise = None
    if 'n_photons' in config and 'n_photons' not in kwargs:
        if method != 'phot':
            raise GalSimConfigError('n_photons is invalid with method != phot')
        if 'max_extra_noise' in config:
            logger.warning(
                "Both 'max_extra_noise' and 'n_photons' are set in config dict, "
                "ignoring 'max_extra_noise'.")
        kwargs['n_photons'] = ParseValue(config, 'n_photons', base, int)[0]
    elif 'max_extra_noise' in config:
        max_extra_noise = ParseValue(config, 'max_extra_noise', base, float)[0]
        if method != 'phot' and max_extra_noise is not None:
            raise GalSimConfigError('max_extra_noise is invalid with method != phot')

    if 'poisson_flux' in config and 'poisson_flux' not in kwargs:
        if method != 'phot':
            raise GalSimConfigError('poisson_flux is invalid with method != phot')
        kwargs['poisson_flux'] = ParseValue(config, 'poisson_flux', base, bool)[0]

    if max_extra_noise is not None and 'max_extra_noise' not in kwargs:
        if max_extra_noise < 0.:
            raise GalSimConfigError("image.max_extra_noise cannot be negative")
        if 'image' in base and 'noise' in base['image']:
            noise_var = CalculateNoiseVariance(base)
        else:
            raise GalSimConfigError("Need to specify noise level when using max_extra_noise")
        if noise_var < 0.:
            raise GalSimConfigError("noise_var calculated to be < 0.")
        max_extra_noise *= noise_var
        kwargs['max_extra_noise'] = max_extra_noise

    bandpass = base.get('bandpass', None)
    if bandpass is not None:
        kwargs['bandpass'] = bandpass
    elif isinstance(prof, ChromaticObject):
        raise GalSimConfigError("Drawing chromatic object requires specifying bandpass")

    if 'photon_ops' in config:
        kwargs['photon_ops'] = BuildPhotonOps(config, 'photon_ops', base, logger)

    if sensor is not None:
        sensor.updateRNG(rng)
        kwargs['sensor'] = sensor

    if image is None:
        kwargs['dtype'] = ParseDType(config, base)

    if logger.isEnabledFor(logging.DEBUG):
        # Don't output the full image array.  Use str(image) for that kwarg.  And Bandpass.
        alt_kwargs = dict([(k, str(kwargs[k]) if isinstance(kwargs[k],(Image,Bandpass))
                               else kwargs[k])
                           for k in kwargs])
        logger.debug('obj %d: drawImage kwargs = %s',base.get('obj_num',0), alt_kwargs)
        logger.debug('obj %d: prof = %s',base.get('obj_num',0),prof)
    try:
        image = prof.drawImage(**kwargs)
    except Exception as e:
        logger.debug('obj %d: prof = %r', base.get('obj_num',0), prof)
        raise
    logger.debug('obj %d: added_flux = %s',base.get('obj_num',0), image.added_flux)
    return image

def ParseWorldPos(config, param_name, base, logger):
    """A helper function to parse the 'world_pos' value.

    The world_pos can be specified either as a regular RA, Dec (which in GalSim is known as a
    CelestialCoord) or as Euclidean coordinates in the local tangent plane relative to the
    image center (a PositionD).

    1. For the RA/Dec option, the world_pos field should use the type RADec, which includes two
       values named ra and dec, each of which should be an Angle type.  e.g.::

            world_pos:
                type : RADec
                ra : 37 hours
                dec: -23 degrees

       Technically, any other type that results in a CelestialCoord is valid, but RADec is the
       only one that is defined natively in GalSim.

    2. For the relative position in the local tangent plane (where 0,0 is the position of the
       image center), you can use any PositionD type.  e.g.::

            world_pos:
                type : RandomCircle
                radius : 12       # arcsec
                inner_radius : 3  # arcsec

    Parameters:
        config:     The configuration dict for the stamp field.
        param_name: The name of the field in the config dict to parse as a world_pos.
                    Normally, this is just 'world_pos'.
        base:       The base configuration dict.

    Returns:
        either a CelestialCoord or a PositionD instance.
    """
    param = config[param_name]
    wcs = base.get('wcs', PixelScale(1.0)) # should be here, but just in case...
    if wcs._isCelestial:
        return ParseValue(config, param_name, base, CelestialCoord)[0]
    else:
        return ParseValue(config, param_name, base, PositionD)[0]

def ParseDType(config, base):
    dtype = config.get('dtype', None)
    if isinstance(dtype, str):
        try:
            try:
                dtype = np.dtype(dtype).type
            except Exception:
                gdict = globals().copy()
                exec('import numpy', gdict)
                exec('import numpy as np', gdict)
                dtype = eval(dtype, gdict)
        except Exception:
            raise GalSimConfigValueError(
                "dtype = %s is invalid."%dtype, Image.valid_dtypes) from None
    if dtype is None and base.get('current_image', None) is not None:
        dtype = base['current_image'].dtype
    return dtype

class StampBuilder:
    """A base class for building stamp images of individual objects.

    The base class defines the call signatures of the methods that any derived class should follow.
    It also includes the implementation of the default stamp type: Basic.
    """

    def setup(self, config, base, xsize, ysize, ignore, logger):
        """
        Do the initialization and setup for building a postage stamp.

        In the base class, we check for and parse the appropriate size and position values in
        config (aka base['stamp'] or base['image'].

        Values given in base['stamp'] take precedence if these are given in both places (which
        would be confusing, so probably shouldn't do that, but there might be a use case where it
        would make sense).

        Parameters:
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            xsize:      The xsize of the image to build (if known).
            ysize:      The ysize of the image to build (if known).
            ignore:     A list of parameters that are allowed to be in config that we can
                        ignore here. i.e. it won't be an error if these parameters are present.
            logger:     A logger object to log progress.

        Returns:
            xsize, ysize, image_pos, world_pos
        """
        # Check for spurious parameters
        CheckAllParams(config, ignore=ignore)

        # Update the size if necessary
        image = base['image']
        if 'xsize' in config:
            xsize = ParseValue(config,'xsize',base,int)[0]
        elif 'size' in config:
            xsize = ParseValue(config,'size',base,int)[0]
        elif 'stamp_xsize' in image:
            xsize = ParseValue(image,'stamp_xsize',base,int)[0]
        elif 'stamp_size' in image:
            xsize = ParseValue(image,'stamp_size',base,int)[0]
        # else use the input xsize

        if 'ysize' in config:
            ysize = ParseValue(config,'ysize',base,int)[0]
        elif 'size' in config:
            ysize = ParseValue(config,'size',base,int)[0]
        elif 'stamp_ysize' in image:
            ysize = ParseValue(image,'stamp_ysize',base,int)[0]
        elif 'stamp_size' in image:
            ysize = ParseValue(image,'stamp_size',base,int)[0]
        # else use the input ysize

        # Determine where this object is going to go:
        if 'image_pos' in config:
            image_pos = ParseValue(config, 'image_pos', base, PositionD)[0]
        elif 'image_pos' in image:
            image_pos = ParseValue(image, 'image_pos', base, PositionD)[0]
        else:
            image_pos = None

        if 'world_pos' in config:
            world_pos = ParseWorldPos(config, 'world_pos', base, logger)
        elif 'world_pos' in image:
            world_pos = ParseWorldPos(image, 'world_pos', base, logger)
        else:
            world_pos = None

        return xsize, ysize, image_pos, world_pos

    def quickSkip(self, config, base):
        """Check whether this object should be skipped before doing any work.

        The base class looks for stamp.quick_skip and returns True if it is preset and
        evaluates to True.

        @param config       The configuration dict for the stamp field.
        @param base         The base configuration dict.

        @returns skip
        """
        return ('quick_skip' in config and ParseValue(config, 'quick_skip', base, bool)[0])

    def setupRNG(self, config, base, logger):
        """Setup the RNG for this object.

        @param config       The configuration dict for the stamp field.
        @param base         The base configuration dict.
        """
        if 'obj_rng' in config:
            if not ParseValue(config,'obj_rng',base,bool)[0]:
                # Just use the image_num rng(s).
                base['obj_num_rngs'] = base.get('image_num_rngs',None)
                base['obj_num_rng'] = base.get('image_num_rng',None)
                return

        # Add 1 to the seed here so the first object has a different rng than the file or image.
        seed = SetupConfigRNG(base, seed_offset=1, logger=logger)
        logger.debug('obj %d: seed = %d',base.get('obj_num',0),seed)

    def locateStamp(self, config, base, xsize, ysize, image_pos, world_pos, logger):
        """Determine where and how large the stamp should be.

        The base class version does the followin:

        - If given, set base['stamp_xsize'] = xsize
        - If given, set base['stamp_ysize'] = ysize
        - If only image_pos or world_pos is given, compute the other from base['wcs']
        - Set base['index_pos'] = image_pos
        - Set base['world_pos'] = world_pos
        - Calculate the appropriate value of the center of the stamp, to be used with the
          command: stamp_image.setCenter(stamp_center).  Save this as base['stamp_center']
        - Calculate the appropriate offset for the position of the object from the center of
          the stamp due to just the fractional part of the image position, not including
          any base['stamp']['offset'] item that may be present in the base dict.
          Save this as base['stamp_offset']

        @param config       The configuration dict for the stamp field.
        @param base         The base configuration dict.
        @param xsize        The size of the stamp in the x-dimension. [may be 0 if unknown]
        @param ysize        The size of the stamp in the y-dimension. [may be 0 if unknown]
        @param image_pos    The position of the stamp in image coordinates. [may be None]
        @param world_pos    The position of the stamp in world coordinates. [may be None]
        @param logger       A logger object to log progress.
        """
        logger = LoggerWrapper(logger)

        # Make sure we have a valid wcs in case image-level processing was skipped.
        if 'wcs' not in base:
            base['wcs'] = BuildWCS(base['image'], 'wcs', config, logger)
        wcs = base['wcs']

        if xsize: base['stamp_xsize'] = xsize
        if ysize: base['stamp_ysize'] = ysize

        # If we have either image_pos or world_pos, calculate the other.
        if image_pos is not None and world_pos is None:
            world_pos = wcs.toWorld(image_pos)
        elif world_pos is not None and image_pos is None:
            image_pos = wcs.toImage(world_pos)

        # If the world_pos is a CelestialCoord, then we also call it sky_pos.
        # If the world_pos is not celestial, or the user wants to override it for some reason,
        # then the user may optionally define a sky_pos value, which gets saved as base['sky_pos'].
        # This may be useful for things that need to know where in the sky the pointing is,
        # even if the WCS is not a CelestialWCS.
        if 'sky_pos' in config:
            base['sky_pos'] = ParseValue(config, 'sky_pos', base, CelestialCoord)[0]
        elif isinstance(world_pos, CelestialCoord):
            base['sky_pos'] = world_pos

        # Sometimes we need a "world" position as flat position, rather than a CelestialCoord.
        # This is called uv_pos.  If the WCS is a EuclideanWCS, then uv_pos = world_pos.
        # If the WCS is a CelestialWCS, then sky_pos = world_pos, and we calculate uv_pos as
        # the tangent-plane projection.
        if isinstance(world_pos, CelestialCoord):
            # Then project this position relative to the image center.
            world_center = base.get('world_center', wcs.toWorld(base['image_center']))
            u, v = world_center.project(world_pos, projection='gnomonic')
            base['uv_pos'] = PositionD(u/arcsec, v/arcsec)
        else:
            base['uv_pos'] = world_pos

        if image_pos is not None:
            # The image_pos refers to the location of the true center of the image, which is
            # not necessarily the nominal center we need for adding to the final image.  In
            # particular, even-sized images have their nominal center offset by 1/2 pixel up
            # and to the right.
            # N.B. This works even if xsize,ysize == 0, since the auto-sizing always produces
            # even sized images.
            nominal_x = image_pos.x        # Make sure we don't change image_pos, which is
            nominal_y = image_pos.y        # stored in base['image_pos'].
            if xsize % 2 == 0: nominal_x += 0.5
            if ysize % 2 == 0: nominal_y += 0.5

            stamp_center = PositionI(int(math.floor(nominal_x+0.5)),
                                     int(math.floor(nominal_y+0.5)))
            stamp_offset = PositionD(nominal_x-stamp_center.x,
                                     nominal_y-stamp_center.y)
        else:
            stamp_center = None
            stamp_offset = PositionD(0.,0.)
            # Set the image_pos to the image center in case the wcs needs it.  Probably, if
            # there is no image_pos or world_pos defined, then it is unlikely a
            # non-trivial wcs will have been set.  So anything would actually be fine.
            image_pos = PositionD( (xsize+1.)/2, (ysize+1.)/2 )

        base['stamp_center'] = stamp_center
        base['stamp_offset'] = stamp_offset
        base['image_pos'] = image_pos
        base['world_pos'] = world_pos

        if xsize:
            logger.debug('obj %d: xsize,ysize = %s,%s',base.get('obj_num',0),xsize,ysize)
        logger.debug('obj %d: image_pos = %s',base.get('obj_num',0),image_pos)
        if world_pos:
            logger.debug('obj %d: world_pos = %s',base.get('obj_num',0),world_pos)
        if stamp_center:
            logger.debug('obj %d: stamp_center = %s',base.get('obj_num',0),stamp_center)

    def getSkip(self, config, base, logger):
        """Initial check of whether to skip this object based on the stamp.skip field.

        @param config       The configuration dict for the stamp field.
        @param base         The base configuration dict.
        @param logger       A logger object to log progress.

        @returns skip
        """
        if 'skip' in config:
            skip = ParseValue(config, 'skip', base, bool)[0]
            if skip:
                logger.debug("obj %d: stamp.skip = True",base.get('obj_num',0))
        else:
            skip = False
        return skip

    def buildPSF(self, config, base, gsparams, logger):
        """Build the PSF object.

        For the Basic stamp type, this builds a PSF from the base['psf'] dict, if present,
        else returns None.

        Parameters:
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            gsparams:   A dict of kwargs to use for a GSParams.  More may be added to this
                        list by the galaxy object.
            logger:     A logger object to log progress.

        Returns:
            the PSF
        """
        return BuildGSObject(base, 'psf', gsparams=gsparams, logger=logger)[0]

    def buildProfile(self, config, base, psf, gsparams, logger):
        """Build the surface brightness profile (a GSObject) to be drawn.

        For the Basic stamp type, this builds a galaxy from the base['gal'] dict and convolves
        it with the psf (if given).  If either the psf or the galaxy is None, then the other one
        is returned as is.

        Parameters:
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            psf:        The PSF, if any.  This may be None, in which case, no PSF is convolved.
            gsparams:   A dict of kwargs to use for a GSParams.  More may be added to this
                        list by the galaxy object.
            logger:     A logger object to log progress.

        Returns:
            the final profile
        """
        gal = BuildGSObject(base, 'gal', gsparams=gsparams, logger=logger)[0]

        if psf:
            if gal:
                return Convolve(gal,psf)
            else:
                return psf
        else:
            if gal:
                return gal
            elif 'gal' in base or 'psf' in base:
                return None
            else:
                raise GalSimConfigError(
                    "At least one of gal or psf must be specified in config. "
                    "If you really don't want any object, use gal type = None.")

    def makeStamp(self, config, base, xsize, ysize, logger):
        """Make the initial empty postage stamp image, if possible.

        If we don't know xsize, ysize, return None, in which case the stamp will be created
        automatically by the drawImage command based on the natural size of the profile.

        Parameters:
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            xsize:      The xsize of the image to build (if known).
            ysize:      The ysize of the image to build (if known).
            logger:     A logger object to log progress.

        Returns:
            the image
        """
        if xsize and ysize:
            dtype = ParseDType(config, base)
            bounds = _BoundsI(1,xsize,1,ysize)

            # Set the origin appropriately
            stamp_center = base['stamp_center']
            if stamp_center:
                bounds = bounds.shift(stamp_center - bounds.center)
            else:
                bounds = bounds.shift(base.get('image_origin',PositionI(1,1)) - PositionI(1,1))

            im = Image(bounds=bounds, dtype=dtype, init_value=0)

            return im
        else:
            return None

    def getDrawMethod(self, config, base, logger):
        """Determine the draw method to use.

        @param config       The configuration dict for the stamp field.
        @param base         The base configuration dict.
        @param logger       A logger object to log progress.

        @returns method
        """
        method = ParseValue(config,'draw_method',base,str)[0]
        if method not in valid_draw_methods:
            raise GalSimConfigValueError("Invalid draw_method.", method, valid_draw_methods)
        return method

    def getOffset(self, config, base, logger):
        """Determine the offset to use.

        The base class version adds the stamp_offset, which comes from calculations related to
        world_pos and image_pos, to the field stamp.offset if any.

        @param config       The configuration dict for the stamp field.
        @param base         The base configuration dict.
        @param logger       A logger object to log progress.

        @returns offset
        """
        offset = base['stamp_offset']
        if 'offset' in config:
            offset += ParseValue(config, 'offset', base, PositionD)[0]
        logger.debug('obj %d: stamp_offset = %s, offset = %s',base.get('obj_num',0),
                     base['stamp_offset'], offset)
        return offset

    def updateSkip(self, prof, image, method, offset, config, base, logger):
        """Before drawing the profile, see whether this object can be trivially skipped.

        The base method checks if the object is completely off the main image, so the
        intersection bounds will be undefined.  In this case, don't bother drawing the
        postage stamp for this object.

        Parameters:
            prof:       The profile to draw.
            image:      The image onto which to draw the profile (which may be None).
            method:     The method to use in drawImage.
            offset:     The offset to apply when drawing.
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            logger:     A logger object to log progress.

        Returns:
            whether to skip drawing this object.
        """
        if isinstance(prof,GSObject) and base.get('current_image',None) is not None:
            if image is None:
                prof = base['wcs'].toImage(prof, image_pos=base['image_pos'])
                prof = prof._shift(-0.5+offset.x, -0.5+offset.y)
                if method in ('auto', 'fft', 'real_space'):
                    prof = Convolution(prof, Pixel(1.))
                N = prof.getGoodImageSize(1.)
                bounds = _BoundsI(1,N,1,N)

                # Set the origin appropriately
                stamp_center = base['stamp_center']
                if stamp_center:
                    bounds = bounds.shift(stamp_center - bounds.center)
                else:
                    bounds = bounds.shift(base.get('image_origin',PositionI(1,1)) - PositionI(1,1))
            else:
                bounds = image.bounds

            overlap = bounds & base['current_image'].bounds
            if not overlap.isDefined():
                logger.info('obj %d: skip drawing object because its image will be entirely off '
                            'the main image.', base.get('obj_num',0))
                return True

        return False

    def draw(self, prof, image, method, offset, config, base, logger):
        """Draw the profile on the postage stamp image.

        Parameters:
            prof:       The profile to draw.
            image:      The image onto which to draw the profile (which may be None).
            method:     The method to use in drawImage.
            offset:     The offset to apply when drawing.
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            logger:     A logger object to log progress.

        Returns:
            the resulting image
        """
        if prof is None:
            return image
        else:
            return DrawBasic(prof,image,method,offset,config,base,logger)

    def whiten(self, prof, image, config, base, logger):
        """If appropriate, whiten the resulting image according to the requested noise profile
        and the amount of noise originally present in the profile.

        Parameters:
            prof:       The profile to draw.
            image:      The image onto which to draw the profile.
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            logger:     A logger object to log progress.

        Returns:
            the variance of the resulting whitened (or symmetrized) image.
        """
        # If the object has a noise attribute, then check if we need to do anything with it.
        current_var = 0.  # Default if not overwritten
        if isinstance(prof,GSObject) and prof.noise is not None:
            if 'image' in base and 'noise' in base['image']:
                noise = base['image']['noise']
                whiten = symmetrize = False
                if 'whiten' in noise:
                    whiten = ParseValue(noise, 'whiten', base, bool)[0]
                if 'symmetrize' in noise:
                    symmetrize = ParseValue(noise, 'symmetrize', base, int)[0]
                if whiten and symmetrize:
                    raise GalSimConfigError('Only one of whiten or symmetrize is allowed')
                if whiten or symmetrize:
                    # In case the galaxy was cached, update the rng
                    rng = GetRNG(noise, base, logger, "whiten")
                    prof.noise.rng.reset(rng)
                if whiten:
                    current_var = prof.noise.whitenImage(image)
                if symmetrize:
                    current_var = prof.noise.symmetrizeImage(image, symmetrize)
        return current_var

    def getSNRScale(self, image, config, base, logger):
        """Calculate the factor by which to rescale the image based on a desired S/N level.

        Note: The default implementation does this for the gal or psf field, so if a custom
              stamp builder uses some other way to get the profiles, this method should
              probably be overridden.

        Parameters:
            image:      The current image.
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            logger:     A logger object to log progress.

        Returns:
            scale_factor
        """
        if 'gal' in base and 'signal_to_noise' in base['gal']:
            key = 'gal'
        elif 'gal' not in base and 'psf' in base and 'signal_to_noise' in base['psf']:
            key = 'psf'
        else:
            return 1.

        if 'flux' in base[key]:
            raise GalSimConfigError(
                'Only one of signal_to_noise or flux may be specified for %s'%key)

        if 'image' in base and 'noise' in base['image']:
            noise_var = CalculateNoiseVariance(base)
        else:
            raise GalSimConfigError(
                "Need to specify noise level when using %s.signal_to_noise"%key)
        sn_target = ParseValue(base[key], 'signal_to_noise', base, float)[0]

        # Now determine what flux we need to get our desired S/N
        # There are lots of definitions of S/N, but here is the one used by Great08
        # We use a weighted integral of the flux:
        # S = sum W(x,y) I(x,y) / sum W(x,y)
        # N^2 = Var(S) = sum W(x,y)^2 Var(I(x,y)) / (sum W(x,y))^2
        # Now we assume that Var(I(x,y)) is dominated by the sky noise, so
        # Var(I(x,y)) = var
        # We also assume that we are using a matched filter for W, so W(x,y) = I(x,y).
        # Then a few things cancel and we find that
        # S/N = sqrt( sum I(x,y)^2 / var )

        sn_meas = np.sqrt( np.sum(image.array**2, dtype=float) / noise_var )
        # Now we rescale the flux to get our desired S/N
        scale_factor = sn_target / sn_meas
        return scale_factor

    def applySNRScale(self, image, prof, scale_factor, method, logger):
        """Apply the scale_factor from getSNRScale to the image and profile.

        The default implementaion just multiplies each of them, but if prof is not a regular
        GSObject, then you might need to do something different.

        Parameters:
            image:          The current image.
            prof:           The profile that was drawn.
            scale_factor:   The factor by which to scale both image and prof.
            method:         The method used by drawImage.
            logger:         A logger object to log progress.

        Returns:
            image, prof  (after being properly scaled)
        """
        if scale_factor != 1.0:
            if method == 'phot':
                logger.warning(
                    "signal_to_noise calculation is not accurate for draw_method = phot")
            image *= scale_factor
            prof *= scale_factor
        return image, prof

    def updateOrigin(self, config, base, image):
        """Update the image origin to be centered at the right place

        @param config       The configuration dict for the stamp field.
        @param base         The base configuration dict.
        @param image        The postage stamp image.
        """
        stamp_center = base['stamp_center']
        if stamp_center:
            image.setCenter(stamp_center)
        else:
            image.setOrigin(base.get('image_origin',PositionI(1,1)))


    def reject(self, config, base, prof, psf, image, logger):
        """Check to see if this object should be rejected.

        Parameters:
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            prof:       The profile that was drawn.
            psf:        The psf that was used to build the profile.
            image:      The postage stamp image.  No noise is on it yet at this point.
            logger:     A logger object to log progress.

        Returns:
            whether to reject this object
        """
        # Early exit if no profile
        if prof is None:
            return False

        if 'reject' in config:
            if ParseValue(config, 'reject', base, bool)[0]:
                logger.info('obj %d: reject evaluated to True',base.get('obj_num',0))
                return True
        if 'min_flux_frac' in config:
            if not isinstance(prof, GSObject):
                raise GalSimConfigError(
                    "Cannot apply min_flux_frac for stamp types that do not use "
                    "a single GSObject profile.")
            expected_flux = prof.flux
            measured_flux = np.sum(image.array, dtype=float)
            min_flux_frac = ParseValue(config, 'min_flux_frac', base, float)[0]
            logger.debug('obj %d: flux_frac = %f', base.get('obj_num',0),
                         measured_flux / expected_flux)
            if measured_flux < min_flux_frac * expected_flux:
                logger.warning('Object %d: Measured flux = %f < %s * %f.',
                               base.get('obj_num',0), measured_flux, min_flux_frac, expected_flux)
                return True
        if 'min_snr' in config or 'max_snr' in config:
            if not isinstance(prof, GSObject):
                raise GalSimConfigError(
                    "Cannot apply min_snr for stamp types that do not use "
                    "a single GSObject profile.")
            var = CalculateNoiseVariance(base)
            sumsq = np.sum(image.array**2, dtype=float)
            snr = np.sqrt(sumsq / var)
            logger.debug('obj %d: snr = %f', base.get('obj_num',0), snr)
            if 'min_snr' in config:
                min_snr = ParseValue(config, 'min_snr', base, float)[0]
                if snr < min_snr:
                    logger.warning('Object %d: Measured snr = %f < %s.',
                                   base.get('obj_num',0), snr, min_snr)
                    return True
            if 'max_snr' in config:
                max_snr = ParseValue(config, 'max_snr', base, float)[0]
                if snr > max_snr:
                    logger.warning('Object %d: Measured snr = %f > %s.',
                                   base.get('obj_num',0), snr, max_snr)
                    return True
        return False

    def reset(self, base, logger):
        """Reset some aspects of the config dict so the object can be rebuilt after rejecting the
        current object.

        Parameters:
            base:       The base configuration dict.
            logger:     A logger object to log progress.
        """
        # Clear current values out of psf, gal, and stamp if they are not safe to reuse.
        # This means they are either marked as safe or indexed by something other than obj_num.
        for field in ('psf', 'gal', 'stamp'):
            if field in base:
                RemoveCurrent(base[field], keep_safe=True, index_key='obj_num')

    def addNoise(self, config, base, image, current_var, logger):
        """
        Add the sky level and the noise to the stamp.

        Note: This only gets called if the image type requests that the noise be added to each
            stamp individually, rather than to the full image and the end.

        Parameters:
            config:         The configuration dict for the stamp field.
            base:           The base configuration dict.
            image:          The current image.
            current_var:    The current noise variance present in the image already.
            logger:         A logger object to log progress.

        Returns:
            the new values of image, current_var
        """
        base['current_noise_image'] = base['current_stamp']
        AddSky(base,image)
        current_var = AddNoise(base,image,current_var,logger)
        return image, current_var

    def makeTasks(self, config, base, jobs, logger):
        """Turn a list of jobs into a list of tasks.

        For the Basic stamp type, there is just one job per task, so the tasks list is just:

            tasks = [ [ (job, k) ] for k, job in enumerate(jobs) ]

        Parameters:
            config:     The configuration dict for the stamp field.
            base:       The base configuration dict.
            jobs:       A list of jobs to split up into tasks.  Each job in the list is a
                        dict of parameters that includes 'obj_num'.
            logger:     A logger object to log progress.

        Returns:
            a list of tasks
        """
        return [ [(job, k)] for k, job in enumerate(jobs) ]


def RegisterStampType(stamp_type, builder):
    """Register an image type for use by the config apparatus.

    Parameters:
        stamp_type:     The name of the type in config['stamp']
        builder:        A builder object to use for building the stamp images.  It should be
                        an instance of StampBuilder or a subclass thereof.
    """
    valid_stamp_types[stamp_type] = builder

RegisterStampType('Basic', StampBuilder())



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'Exponential' ]

import math

from . import _galsim
from .gsobject import GSObject
from .gsparams import GSParams
from .utilities import lazy_property, doc_inherit
from .errors import GalSimIncompatibleValuesError


class Exponential(GSObject):
    r"""A class describing an exponential profile.

    Surface brightness profile with

    .. math::
        I(r) \sim e^{-r/r_0}

    where :math:`r_0` is the ``scale_radius``.  This is a special case of the `Sersic` profile,
    but is given a separate class since the Fourier transform has closed form and can be generated
    without lookup tables.

    An Exponential can be initialized using one (and only one) of two possible size parameters:
    ``scale_radius`` or ``half_light_radius``.  Exactly one of these two is required.

    Parameters:
        half_light_radius:  The half-light radius of the profile.  Typically given in arcsec.
                            [One of ``scale_radius`` or ``half_light_radius`` is required.]
        scale_radius:       The scale radius of the profile.  Typically given in arcsec.
                            [One of ``scale_radius`` or ``half_light_radius`` is required.]
        flux:               The flux (in photons/cm^2/s) of the profile. [default: 1]
        gsparams:           An optional `GSParams` argument. [default: None]
    """
    _opt_params = { "flux" : float }
    _single_params = [ { "scale_radius" : float , "half_light_radius" : float } ]

    # The half-light-radius is not analytic, but can be calculated numerically
    # by iterative solution of equation:
    #     (re / r0) = ln[(re / r0) + 1] + ln(2)
    _hlr_factor = 1.6783469900166605
    _one_third = 1./3.
    _inv_twopi = 0.15915494309189535

    _has_hard_edges = False
    _is_axisymmetric = True
    _is_analytic_x = True
    _is_analytic_k = True

    def __init__(self, half_light_radius=None, scale_radius=None, flux=1., gsparams=None):
        if half_light_radius is not None:
            if scale_radius is not None:
                raise GalSimIncompatibleValuesError(
                    "Only one of scale_radius and half_light_radius may be specified",
                    half_light_radius=half_light_radius, scale_radius=scale_radius)
            else:
                scale_radius = half_light_radius / Exponential._hlr_factor
        elif scale_radius is None:
                raise GalSimIncompatibleValuesError(
                    "Either scale_radius or half_light_radius must be specified",
                    half_light_radius=half_light_radius, scale_radius=scale_radius)
        self._r0 = float(scale_radius)
        self._flux = float(flux)
        self._gsparams = GSParams.check(gsparams)
        self._inv_r0 = 1./self._r0
        self._norm = self._flux * Exponential._inv_twopi * self._inv_r0**2

    @lazy_property
    def _sbp(self):
        return _galsim.SBExponential(self._r0, self._flux, self.gsparams._gsp)

    @property
    def scale_radius(self):
        """The scale radius of the profile.
        """
        return self._r0
    @property
    def half_light_radius(self):
        """The half-light radius of the profile.
        """
        return self._r0 * Exponential._hlr_factor

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, Exponential) and
                 self.scale_radius == other.scale_radius and
                 self.flux == other.flux and
                 self.gsparams == other.gsparams))

    def __hash__(self):
        return hash(("galsim.Exponential", self.scale_radius, self.flux, self.gsparams))

    def __repr__(self):
        return 'galsim.Exponential(scale_radius=%r, flux=%r, gsparams=%r)'%(
            self.scale_radius, self.flux, self.gsparams)

    def __str__(self):
        s = 'galsim.Exponential(scale_radius=%s'%self.scale_radius
        if self.flux != 1.0:
            s += ', flux=%s'%self.flux
        s += ')'
        return s

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_sbp',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d

    @property
    def _maxk(self):
        return (self.gsparams.maxk_threshold ** -Exponential._one_third) / self.scale_radius

    @property
    def _stepk(self):
        return self._sbp.stepK()

    @property
    def _max_sb(self):
        return self._norm

    def _xValue(self, pos):
        r = math.sqrt(pos.x**2 + pos.y**2)
        return self._norm * math.exp(-r * self._inv_r0)

    def _kValue(self, kpos):
        ksqp1 = (kpos.x**2 + kpos.y**2) * self._r0**2 + 1.
        return self._flux / (ksqp1 * math.sqrt(ksqp1))

    def _drawReal(self, image, jac=None, offset=(0.,0.), flux_scaling=1.):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        dx,dy = offset
        self._sbp.draw(image._image, image.scale, _jac, dx, dy, flux_scaling)

    def _shoot(self, photons, rng):
        self._sbp.shoot(photons._pa, rng._rng)

    def _drawKImage(self, image, jac=None):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        self._sbp.drawK(image._image, image.scale, _jac)

    @doc_inherit
    def withFlux(self, flux):
        return Exponential(scale_radius=self.scale_radius, flux=flux, gsparams=self.gsparams)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'SED', 'EmissionLine' ]

import numpy as np
from astropy import units
from astropy import constants
from numbers import Real
from pathlib import PosixPath

from .table import LookupTable, _LookupTable
from ._utilities import WeakMethod, lazy_property, basestring
from .errors import GalSimError, GalSimValueError, GalSimRangeError, GalSimSEDError
from .errors import GalSimIncompatibleValuesError
from .random import DistDeviate
from . import utilities
from . import integ
from . import dcr
from . import gsobject

class SED:
    """Object to represent the spectral energy distributions of stars and galaxies.

    SEDs are callable, usually returning the flux density in photons/nm/cm^2/s as a function of
    wavelength, though SEDs are also used by GalSim to track dimensionless wavelength-dependent
    normalizations, and may thus also return dimensionless values.  By default, the above wavelength
    used by __call__ is nanometers, but it's possible to use other units via the astropy.units
    module (at least, if the SED keyword argument ``fast=False``, see below).  For instance,::

        >>> sed = galsim.SED(...)
        >>> from astropy import units as u
        >>> assert sed(500) == sed(5000 * u.AA)  # 500 nm == 5000 Angstroms

    The python type of the return value depends on the type of the input wavelength(s).  A scalar
    input wavelength yields a scalar flux density, a tuple yields a tuple, a list yields a list, and
    a numpy.ndarray yields a numpy.ndarray.  A scalar astropy.units.Quantity yields a python scalar,
    and a vector astropy.units.Quantity yields a numpy.ndarray.

    SEDs are immutable; all transformative SED methods return *new* SEDs, and leave their
    originating SEDs unaltered.

    SEDs have ``blue_limit`` and ``red_limit`` attributes, which indicate the range over which the
    SED is defined.  An exception will be raised if the flux density or normalization is requested
    outside of this range.  Note that ``blue_limit`` and ``red_limit`` are always in nanometers and
    in the observed frame when ``redshift != 0``.

    SEDs may be multiplied by scalars or scalar functions of wavelength.  In particular, an SED
    multiplied by a `Bandpass` will yield the appropriately filtered SED.  Two SEDs may be
    multiplied together if at least one of them represents a dimensionless normalization.

    SEDs may be added together if they are at the same redshift.  The resulting SED will only be
    defined on the wavelength region where both of the operand SEDs are defined. ``blue_limit`` and
    ``red_limit`` will be reset accordingly.

    The input parameter, ``spec``, may be one of several possible forms:

    1. a regular python function (or an object that acts like a function)
    2. a `LookupTable`
    3. a file from which a `LookupTable` can be read in
    4. a string which can be evaluated to a function of ``wave`` via ``eval('lambda wave:'+spec)``,
       e.g.::

            spec = '0.8 + 0.2 * (wave-800)'

    5. a python scalar (only possible for dimensionless SEDs)

    The argument of ``spec`` should be the wavelength in units specified by ``wave_type``, which
    should be an instance of ``astropy.units.Unit`` of equivalency class ``astropy.units.spectral``,
    or one of the case-insensitive aliases 'nm', 'nanometer', 'nanometers', 'A', 'Ang', 'Angstrom',
    or 'Angstroms'.  Note that ``astropy.units.spectral`` includes not only units with dimensions
    of length, but also frequency, energy, or wavenumber.

    The return value of ``spec`` should be a spectral density with units specified by ``flux_type``,
    which should be an instance of ``astropy.units.Unit`` of equivalency class
    ``astropy.units.spectral_density``, or one of the case-insensitive aliases:

    1. 'flambda':  erg/wave_type/cm^2/s, where wave_type is as above.
    2. 'fnu':      erg/Hz/cm^2/s
    3. 'fphotons': photons/wave_type/cm^2/s, where wave_type is as above.
    4. '1':        dimensionless

    Note that the ``astropy.units.spectral_density`` class includes units with dimensions of
    [energy/time/area/unit-wavelength], [energy/time/area/unit-frequency],
    [photons/time/area/unit-wavelength], and so on.

    Finally, the optional ``fast`` keyword option is used to specify when unit and dimension changes
    are executed, particularly for SEDs specified by a `LookupTable`.  If ``fast=True``, the
    default, then the input units/dimensions may be converted to an internal working unit before
    interpolation in wavelength is performed.  Alternatively, ``fast=False`` implies that
    interpolation should take place in the native units of the input ``spec``, and subsequently flux
    density converted to photons/cm^2/s/nm afterwards.  Generally, the former option is faster, but
    may be less accurate since interpolation and dimensionality conversion do not commute.  One
    consequence of using ``fast=True`` is that __call__ can not accept an ``astropy.units.Quantity``
    in this case.

    Parameters:
        spec:           Function defining the z=0 spectrum at each wavelength.  See above for
                        valid options for this parameter.
        wave_type:      String or astropy.unit specifying units for wavelength input to ``spec``.
        flux_type:      String or astropy.unit specifying what type of spectral density or
                        dimensionless normalization ``spec`` represents.  See above for valid
                        options for this parameter.
        redshift:       Optionally shift the spectrum to the given redshift. [default: 0]
        fast:           Convert units on initialization instead of on __call__. [default: True]
        interpolant:    If reading from a file, what interpolant to use. [default: 'linear']
    """
    # We'll use these multiple times below, and they are ridiculously slow to construct,
    # so just make them once at the class level.
    _fphotons_base = units.astrophys.photon/(units.s * units.cm**2)
    _flambda_base = units.erg/(units.s * units.cm**2)
    _fphotons = _fphotons_base / units.nm
    _flambda = _flambda_base / units.nm
    _fnu = units.erg / (units.s * units.Hz * units.cm**2)
    _spec_nm = units.spectral_density(1*units.nm)
    _c = constants.c.to('nm/s').value
    _h = constants.h.to('erg s').value
    _dimensionless = units.dimensionless_unscaled
    _bolo_max_wave = 1.e30  # What we use as "infinity" for bolometric flux calculations.

    def __init__(self, spec, wave_type, flux_type, redshift=0., fast=True, interpolant='linear',
                 _blue_limit=0.0, _red_limit=np.inf, _wave_list=None):
        self._flux_type = flux_type  # Need to save the original for repr
        self.interpolant = interpolant

        # Parse the various options for wave_type
        self.wave_type, self.wave_factor = self._parse_wave_type(wave_type)

        # Parse the various options for flux_type
        self.flux_factor = None
        if isinstance(flux_type, str):
            if flux_type.lower() == 'flambda':
                self.flux_type = 'flambda'
                self.spectral = True
                self.flux_factor = 1. / (SED._h * SED._c)
            elif flux_type.lower() == 'fphotons':
                self.spectral = True
                if self.wave_factor is not None:
                    self.flux_type = 'fphotons'
                    self.flux_factor = self.wave_factor
                else:
                    self.flux_type = SED._fphotons
            elif flux_type.lower() == 'fnu':
                self.spectral = True
                if self.wave_factor is not None:
                    self.flux_type = 'fnu'
                    self.flux_factor = self.wave_factor / SED._h
                else:
                    self.flux_type = SED._fnu
            elif flux_type == '1':
                self.flux_type = '1'
                self.spectral = False
            else:
                raise GalSimValueError("Unknown flux_type", flux_type,
                                       ('flambda', 'fnu', 'fphotons', '1'))
        else:
            self.flux_type = flux_type
            self.spectral = self.check_spectral()
            if not self.spectral and not self.check_dimensionless():
                raise GalSimValueError(
                    "Flux_type must be equivalent to a spectral density or dimensionless.",
                    flux_type)
            try:
                if self.wave_factor and self.spectral:
                    self.flux_factor = (1*self.flux_type).to(SED._fphotons).value
                    self.flux_type = 'fphotons'
            except units.UnitConversionError:
                try:
                    self.flux_factor = (1*self.flux_type).to(SED._flambda).value
                    self.flux_factor /= SED._h * SED._c * self.wave_factor
                    self.flux_type = 'flambda'
                except units.UnitConversionError:
                    try:
                        self.flux_factor = (1*self.flux_type).to(SED._fnu).value
                        self.flux_factor *= self.wave_factor / SED._h
                        self.flux_type = 'fnu'
                    except units.UnitConversionError:
                        self.wave_type = units.Unit(self.wave_type)

        self.redshift = redshift
        self.fast = fast

        # Convert string input into a real function (possibly a LookupTable)
        self._orig_spec = spec  # Save this for pickling
        self._initialize_spec()

        # Setup the wave_list, red_limit, blue_limit
        if _wave_list is not None:
            self.wave_list = _wave_list
            self.blue_limit = float(_blue_limit)
            self.red_limit = float(_red_limit)
        elif isinstance(self._spec, LookupTable):
            self.wave_list = np.array(self._spec.getArgs())
            if self.wave_factor:
                self.wave_list *= (1.0 + self.redshift) / self.wave_factor
            else:
                self.wave_list = (self.wave_list*self.wave_type).to(units.nm, units.spectral()).value
                self.wave_list *= (1.0 + self.redshift)
            self.blue_limit = float(np.min(self.wave_list))
            self.red_limit = float(np.max(self.wave_list))
        else:
            self.blue_limit = 0.0
            self.red_limit = np.inf
            self.wave_list = np.array([], dtype=float)

        # Define the appropriate functions to call
        self._setup_funcs()

    @staticmethod
    def _parse_wave_type(wave_type):
        # Parse the various options for wave_type.
        # Returns wave_type, wave_factor
        if isinstance(wave_type, str):
            if wave_type.lower() in ('nm', 'nanometer', 'nanometers'):
                return 'nm', 1.
            elif wave_type.lower() in ('a', 'ang', 'angstrom', 'angstroms'):
                return 'Angstrom', 10.
            else:
                raise GalSimValueError("Unknown wave_type", wave_type, ('nm', 'Angstrom'))
        else:
            try:
                wave_factor = (1*units.nm).to(wave_type).value
                if wave_factor == 1.:
                    return 'nm', 1.
                elif abs(wave_factor-10.) < 2.e-15:  # This doesn't come out exactly 10.
                    return 'Angstrom', 10.
                else:
                    return units.Unit(wave_type), wave_factor
            except units.UnitConversionError:
                return units.Unit(wave_type), None

    def _setup_funcs(self):
        # Set up the various functions we use to do the right calculation based on which
        # wave type and/or flux type we have for _spec.
        # The astropy unit functions are horribly slow, so we want to avoid them as much as
        # possible.  If the wave_type and flux_type are one of the simpler (and most common)
        # types, then we have custom functions that do the necessary conversions directly.
        if self.wave_factor == 1:
            self._get_native_waves = WeakMethod(self._get_native_waves_trivial)
        elif self.wave_factor:
            self._get_native_waves = WeakMethod(self._get_native_waves_fast)
        else:
            self._get_native_waves = WeakMethod(self._get_native_waves_slow)

        if self.redshift == 0.:
            self._get_rest_native_waves = self._get_native_waves
        elif self.wave_factor:
            self._get_rest_native_waves = WeakMethod(self._get_rest_native_waves_fast)
        else:
            self._get_rest_native_waves = WeakMethod(self._get_rest_native_waves_slow)

        if self.flux_type == 'fphotons':
            #assert self.flux_factor is not None
            self._flux_to_photons = WeakMethod(self._flux_to_photons_fphot)
        elif self.flux_type == 'flambda':
            #assert self.flux_factor is not None
            self._flux_to_photons = WeakMethod(self._flux_to_photons_flam)
        elif self.flux_type == 'fnu':
            #assert self.flux_factor is not None
            self._flux_to_photons = WeakMethod(self._flux_to_photons_fnu)
        else:
            self._flux_to_photons = WeakMethod(self._flux_to_photons_slow)

        if self.fast:
            self._call = WeakMethod(self._call_fast)
        else:
            self._call = WeakMethod(self._call_slow)

    # Here are the definitions for the various functions we can use depending on the wave_type
    # and flux_type (cf. _setup_funcs).
    def _get_native_waves_trivial(self, wave):
        return wave

    def _get_native_waves_fast(self, wave):
        return np.asarray(wave) * self.wave_factor

    def _get_native_waves_slow(self, wave):
        return (wave * units.nm).to(self.wave_type, units.spectral()).value

    def _get_rest_native_waves_fast(self, wave):
        return np.asarray(wave) * (self.wave_factor / (1.0+self.redshift))

    def _get_rest_native_waves_slow(self, wave):
        return (wave / (1.0+self.redshift) * units.nm).to(self.wave_type, units.spectral()).value

    def _flux_to_photons_fphot(self, flux_native, wave_native):
        return flux_native * self.flux_factor

    def _flux_to_photons_flam(self, flux_native, wave_native):
        return flux_native * wave_native * self.flux_factor

    def _flux_to_photons_fnu(self, flux_native, wave_native):
        return flux_native / wave_native * self.flux_factor

    def _flux_to_photons_slow(self, flux_native, wave_native):
        return (flux_native * self.flux_type).to(
                SED._fphotons, units.spectral_density(wave_native * self.wave_type)).value


    def _initialize_spec(self):
        # Turn the input _orig_spec into a real function _spec.
        # The function cannot be pickled, so will need to do this in getstate as well as init.

        self._const = False
        if isinstance(self._orig_spec, (int, float)):
            if not self.dimensionless:
                raise GalSimSEDError("Attempt to set spectral SED using float or integer.", self)
            self._const = True
            self._spec = lambda w: float(self._orig_spec)
        elif isinstance(self._orig_spec, (basestring, PosixPath)):
            isfile, filename = utilities.check_share_file(self._orig_spec, 'SEDs')
            if isfile:
                self._spec = LookupTable.from_file(filename, interpolant=self.interpolant)
            else:
                # If a constant function is input as a string (e.g. '1'), then we want to
                # make sure it is flagged as a const SED.
                try:
                    float(self._orig_spec)
                except ValueError:
                    pass
                else:
                    self._const = True
                    self._spec = lambda w: float(self._orig_spec)
                    return

                # Don't catch ArithmeticErrors when testing to see if the the result of `eval()`
                # is valid since `spec = '1./(wave-700)'` will generate a ZeroDivisionError (which
                # is a subclass of ArithmeticError) despite being a valid spectrum specification,
                # while `spec = 'blah'` where `blah` is undefined generates a NameError and is not
                # a valid spectrum specification.
                # Are there any other types of errors we should trap here?
                try:
                    self._spec = utilities.math_eval('lambda wave : ' + self._orig_spec)
                    test_value = self._spec(700.0)
                except ArithmeticError:
                    test_value = 0
                except Exception as e:
                    raise GalSimValueError(
                        "String spec must either be a valid filename or something that "
                        "can eval to a function of wave.\n"
                        "Caught error: {0}".format(e), self._orig_spec)
                if not isinstance(test_value, Real):
                    raise GalSimValueError("The given SED function did not return a valid number "
                                           "at test wavelength %s: got %s"%(700.0, test_value),
                                           self._orig_spec)

        else:
            self._spec = self._orig_spec

    def check_spectral(self):
        """Return boolean indicating if SED has units compatible with a spectral density."""
        return self.flux_type.is_equivalent(SED._fphotons, SED._spec_nm)

    def check_dimensionless(self):
        """Return boolean indicating if SED is dimensionless."""
        if self.flux_type.is_equivalent(SED._dimensionless):
            self._flux_type = '1'
            # The astropy.units.dimensionless_unscaled object isn't properly reprable.
            # So switch to using '1' in these cases.
            return True
        else:
            return False

    @property
    def dimensionless(self):  # for convenience
        """Whether the object is dimensionless (rather than spectral).
        """
        return not self.spectral

    def _rest_nm_to_photons(self, wave):
        wave_native = self._get_native_waves(wave)
        flux_native = self._spec(wave_native)
        return self._flux_to_photons(flux_native, wave_native)

    def _rest_nm_to_dimensionless(self, wave):
        return self._spec(self._get_native_waves(wave))

    def _check_bounds(self, wave):
        if hasattr(wave, '__iter__'):
            wmin = np.min(wave)
            wmax = np.max(wave)
        else:
            wmin = wmax = wave

        extrapolation_slop = 1.e-6 # allow a small amount of extrapolation
        if wmin < self.blue_limit - extrapolation_slop:
            raise GalSimRangeError("Requested wavelength is bluer than blue_limit.",
                                   wave, self.blue_limit, self.red_limit)
        if wmax > self.red_limit + extrapolation_slop:
            raise GalSimRangeError("Requested wavelength is redder than red_limit.",
                                   wave, self.blue_limit, self.red_limit)

    @lazy_property
    def _fast_spec(self):
        # Create a fast version of self._spec by constructing a LookupTable on self.wave_list
        if self.wave_factor == 1. and self.flux_factor == 1.:
            return self._spec
        else:
            if len(self.wave_list) == 0:
                if self.spectral:
                    return WeakMethod(self._rest_nm_to_photons)
                else:
                    return WeakMethod(self._rest_nm_to_dimensionless)
            else:
                x = self.wave_list / (1.0 + self.redshift)
                if self.spectral:
                    f = self._rest_nm_to_photons(x)
                else:
                    f = self._rest_nm_to_dimensionless(x)
                interp = self._spec.interpolant if isinstance(self._spec, LookupTable) else 'linear'
                return _LookupTable(x, f, interpolant=interp)

    def _call_fast(self, wave):
        """Return either flux in photons / sec / cm^2 / nm, or dimensionless normalization.

        Assumes that self._spec has already been transformed to accept correct wavelength units and
        yield correct flux units.

        Parameters:
            wave:   Wavelength in nanometers.

        Returns:
            Flux or normalization.
        """
        self._check_bounds(wave)
        return self._fast_spec(np.asarray(wave) / (1.0 + self.redshift))

    def _call_slow(self, wave):
        """Return flux in photons / sec / cm^2 / nm or dimensionless normalization.

        Uses self._spec that has not been pre-transformed for desired units, instead does all unit
        conversions inside this method.

        Parameters:
            wave:   Wavelength.  If not an astropy.units.Quantity, then assumed units are
                    nanometers.

        Returns:
            Flux.
        """
        wave_in = wave
        # Convert wave to nanometers if needed.
        if isinstance(wave, units.Quantity):
            wave = wave.to(units.nm, units.spectral()).value

        wave = np.asarray(wave)

        self._check_bounds(wave)

        # Figure out rest-frame wave_type wavelength array for query to self._spec.
        rest_wave_native = self._get_rest_native_waves(wave)

        out = self._spec(rest_wave_native)

        # Manipulate output units
        if self.spectral:
            out = self._flux_to_photons(out, rest_wave_native)
        return out

    def __call__(self, wave):
        """Return photon flux density or dimensionless normalization at wavelength ``wave``.

        Note that outside of the wavelength range defined by the ``blue_limit`` and ``red_limit``
        attributes, the SED is considered undefined, and this method will raise an exception if a
        wavelength outside the defined range is passed as an argument.

        Parameters:
            wave:   Wavelength in nanometers at which to evaluate the SED. May be a scalar,
                    a numpy.array, or an astropy.units.Quantity

        Returns:
            photon flux density in units of photons/nm/cm^2/s if self.spectral, or
            dimensionless normalization if self.dimensionless.
        """
        ret = self._call(wave)
        if self._const:
            ret = np.full_like(wave, ret, dtype=float)
        return ret

    def _mul_sed(self, other):
        """Equivalent to self * other when other is an SED, but no sanity checks."""
        # There should only be one SED with a non-trivial redshift, so adding them
        # should always give us the right net redshift to use.
        redshift = self.redshift + other.redshift

        fast = self.fast and other.fast

        wave_list, blue_limit, red_limit = utilities.combine_wave_list(self, other)
        if fast:
            if (isinstance(self._fast_spec, LookupTable)
                    and not self._fast_spec.x_log
                    and not self._fast_spec.f_log):
                x = wave_list / (1.0 + self.redshift)
                # Add in 500 uniformly spaced values to help improve accuracy.
                x = utilities.merge_sorted([x, np.linspace(x[0], x[-1], 500)])
                zfactor2 = (1.+redshift) / (1.+other.redshift)
                f = self._fast_spec(x) * other._fast_spec(x*zfactor2)
                spec = _LookupTable(x, f, self._fast_spec.interpolant)
            elif (isinstance(other._fast_spec, LookupTable)
                    and not other._fast_spec.x_log
                    and not other._fast_spec.f_log):
                x = wave_list / (1.0 + other.redshift)
                x = utilities.merge_sorted([x, np.linspace(x[0], x[-1], 500)])
                zfactor1 = (1.+redshift) / (1.+other.redshift)
                f = self._fast_spec(x*zfactor1) * other._fast_spec(x)
                spec = _LookupTable(x, f, other._fast_spec.interpolant)
            else:
                zfactor1 = (1.+redshift) / (1.+self.redshift)
                zfactor2 = (1.+redshift) / (1.+other.redshift)
                spec = lambda w: self._fast_spec(w * zfactor1) * other._fast_spec(w * zfactor2)
        else:
            spec = lambda w: self(w * (1.+redshift)) * other(w * (1.+redshift))
        spectral = self.spectral or other.spectral
        flux_type = 'fphotons' if spectral else '1'
        return SED(spec, 'nm', flux_type, redshift=redshift, fast=fast,
                   _blue_limit=blue_limit, _red_limit=red_limit, _wave_list=wave_list)

    def _mul_bandpass(self, other):
        """Equivalent to self * other when other is a Bandpass"""
        wave_list, blue_limit, red_limit = utilities.combine_wave_list(self, other)
        zfactor = (1.0+self.redshift) / other.wave_factor
        if self.fast:
            if (isinstance(self._fast_spec, LookupTable)
                    and not self._fast_spec.x_log
                    and not self._fast_spec.f_log):
                x = wave_list / (1.0 + self.redshift)
                # Add in 500 uniformly spaced values to help improve accuracy.
                x = utilities.merge_sorted([x, np.linspace(x[0], x[-1], 500)])
                f = self._fast_spec(x) * other._tp(x*zfactor)
                spec = _LookupTable(x, f, self._fast_spec.interpolant)
            else:
                spec = lambda w: self._fast_spec(w) * other._tp(w*zfactor)
        else:
            spec = lambda w: self(w*(1.0+self.redshift)) * other._tp(w*zfactor)
        return SED(spec, 'nm', 'fphotons', redshift=self.redshift, fast=self.fast,
                   _blue_limit=blue_limit, _red_limit=red_limit, _wave_list=wave_list)


    def _mul_scalar(self, other, spectral):
        """Equivalent to self * other when other is a scalar"""
        # If other is a scalar and self._spec a LookupTable, then remake that LookupTable.
        if isinstance(self._spec, LookupTable):
            wave_type = self.wave_type
            flux_type = self._flux_type
            x = self._spec.getArgs()
            f = np.array(self._spec.getVals()) * other
            spec = _LookupTable(x, f, x_log=self._spec.x_log, f_log=self._spec.f_log,
                                interpolant=self._spec.interpolant)
        elif self._const and not spectral:
            spec = self._spec(42.0) * other
            wave_type = 'nm'
            flux_type = '1'
        else:
            wave_type = 'nm'
            flux_type = 'fphotons' if spectral else '1'
            if self.fast:
                spec = lambda w: self._fast_spec(w) * other
            else:
                spec = lambda w: self(w*(1.0+self.redshift)) * other
        return SED(spec, wave_type, flux_type, redshift=self.redshift, fast=self.fast,
                   _blue_limit=self.blue_limit, _red_limit=self.red_limit,
                   _wave_list=self.wave_list)


    def __mul__(self, other):
        """Multiply the SED by something.

        There are several possibilities:

        1. SED * SED -> SED (at least one must be dimensionless)
        2. SED * GSObject -> ChromaticObject
        3. SED * Bandpass -> SED (treating throughput similarly to dimensionless SED)
        4. SED * callable function -> SED (treating function as dimensionless SED)
        5. SED * scalar -> SED
        """
        # Watch out for 5 types of `other`:
        # 1.  SED: Check that not both spectral densities.
        # 2.  GSObject: return a ChromaticObject().
        # 3.  Bandpass: return an SED, but carefully propagate blue/red limit and wave_list.
        # 4.  Callable: return an SED
        # 5.  Scalar: return an SED
        #
        # Additionally, check for shortcuts when self._const

        # Product of two SEDs
        if isinstance(other, SED):
            if self.spectral and other.spectral:
                raise GalSimIncompatibleValuesError(
                    "Cannot multiply two spectral densities together.", self_sed=self, other=other)

            if other._const:
                # const, so can eval anywhere.
                return self._mul_scalar(other._spec(42.0), self.spectral or other.spectral)
            elif self._const:
                return other._mul_scalar(self._spec(42.0), self.spectral or other.spectral)
            else:
                return self._mul_sed(other)

        # Product of SED and achromatic GSObject is a SimpleChromaticTransformation.
        elif isinstance(other, gsobject.GSObject):
            return other * self

        # Product of SED and Bandpass is (filtered) SED.  The `redshift` attribute is retained.
        elif isinstance(other, Bandpass):
            return self._mul_bandpass(other)

        # Product of SED with generic callable is also a (filtered) SED, with retained `redshift`.
        elif hasattr(other, '__call__'):
            if self.fast:
                spec = lambda w: self._fast_spec(w) * other(w*(1.0+self.redshift))
            else:
                spec = lambda w: self(w*(1.0+self.redshift)) * other(w*(1.0+self.redshift))
            flux_type = 'fphotons' if self.spectral else '1'
            return SED(spec, 'nm', flux_type, redshift=self.redshift, fast=self.fast,
                       _blue_limit=self.blue_limit, _red_limit=self.red_limit,
                       _wave_list=self.wave_list)

        elif isinstance(other, (int, float)):
            return self._mul_scalar(other, self.spectral)

        else:
            raise TypeError("Cannot multiply an SED by %s"%(other))

    def __rmul__(self, other):
        return self*other

    def __div__(self, other):
        # Enable division by scalars or dimensionless callables (including dimensionless SEDs.)
        if isinstance(other, SED) and other.spectral:
            raise GalSimSEDError("Cannot divide by spectral SED.", other)
        if hasattr(other, '__call__'):
            spec = lambda w: self(w * (1.0 + self.redshift)) / other(w * (1.0 + self.redshift))
        elif isinstance(self._spec, LookupTable):
            # If other is not a function, then there is no loss of accuracy by applying the
            # factor directly to the LookupTable, if that's what we are using.
            # Make sure to keep the same properties about the table, flux_type, wave_type.
            x = self._spec.getArgs()
            f = [ val / other for val in self._spec.getVals() ]
            spec = _LookupTable(x, f, x_log=self._spec.x_log, f_log=self._spec.f_log,
                                interpolant=self._spec.interpolant)
        else:
            spec = lambda w: self(w * (1.0 + self.redshift)) / other

        return SED(spec, flux_type=self.flux_type, wave_type=self.wave_type,
                   redshift=self.redshift, fast=self.fast,
                   _wave_list=self.wave_list,
                   _blue_limit=self.blue_limit, _red_limit=self.red_limit)

    __truediv__ = __div__

    def __add__(self, other):
        # Add together two SEDs, with the following caveats:
        # 1) The SEDs must have the same redshift.
        # 2) The resulting SED will be defined on the wavelength range set by the overlap of the
        #    wavelength ranges of the two SED operands.
        # 3) The new `wave_list` will be the union of the operand `wave_list`s in the intersecting
        #    region, even if one or both of the `wave_list`s are empty.
        # These conditions ensure that SED addition is commutative.

        if self.redshift != other.redshift:
            raise GalSimIncompatibleValuesError(
                "Can only add SEDs with same redshift.", self_sed=self, other=other)

        if self.dimensionless and other.dimensionless:
            flux_type = '1'
        elif self.spectral and other.spectral:
            flux_type = 'fphotons'
        else:
            raise GalSimIncompatibleValuesError(
                "Cannot add SEDs with incompatible dimensions.", self_sed=self, other=other)

        wave_list, blue_limit, red_limit = utilities.combine_wave_list(self, other)

        # If both SEDs are `fast`, and both `_fast_spec`s are LookupTables, then make a new
        # LookupTable instead and preserve picklability.
        # First need to make sure self._fast_spec and other._fast_spec are initialized.  Can do this
        # by evaluating them at a good wavelength.  blue_limit should work.
        self(blue_limit)
        other(blue_limit)
        if (self.fast
                and other.fast
                and isinstance(self._fast_spec, LookupTable)
                and isinstance(other._fast_spec, LookupTable)
                and not self._fast_spec.x_log
                and not other._fast_spec.x_log
                and not self._fast_spec.f_log
                and not other._fast_spec.f_log
                and self._fast_spec.interpolant == 'linear'
                and other._fast_spec.interpolant == 'linear'):
            x = wave_list / (1.0 + self.redshift)
            f = self._fast_spec(x) + other._fast_spec(x)
            # Note: adding splines doesn't quite work at full precision, so only do this for
            # linear interpolants.
            spec = _LookupTable(x, f, interpolant='linear')
        else:
            spec = lambda w: self(w*(1.0+self.redshift)) + other(w*(1.0+self.redshift))

        return SED(spec, wave_type='nm', flux_type=flux_type,
                   redshift=self.redshift, fast=self.fast, _wave_list=wave_list,
                   _blue_limit=blue_limit, _red_limit=red_limit)

    def __sub__(self, other):
        # Subtract two SEDs, with the same caveats as adding two SEDs.
        return self.__add__(-1.0 * other)

    def withFluxDensity(self, target_flux_density, wavelength):
        """Return a new `SED` with flux density set to ``target_flux_density`` at wavelength
        ``wavelength``.

        See `ChromaticObject` docstring for information about how `SED` normalization affects
        `ChromaticObject` normalization.

        Parameters:
            target_flux_density:    The target normalization in photons/nm/cm^2/s.
            wavelength:             The wavelength, in nm, at which the flux density will be set.

        Returns:
            the new normalized SED.
        """
        if self.dimensionless:
            raise GalSimSEDError("Cannot set flux density of dimensionless SED.", self)
        if isinstance(wavelength, units.Quantity):
            wavelength_nm = wavelength.to(units.nm, units.spectral())
            current_flux_density = self._call(wavelength_nm.value)
        else:
            wavelength_nm = wavelength * units.nm
            current_flux_density = self._call(wavelength)
        if isinstance(target_flux_density, units.Quantity):
            target_flux_density = target_flux_density.to(
                    SED._fphotons, units.spectral_density(wavelength_nm)).value
        factor = target_flux_density / current_flux_density
        return self * factor

    def withFlux(self, target_flux, bandpass):
        """Return a new `SED` with flux through the `Bandpass` ``bandpass`` set to ``target_flux``.

        See `ChromaticObject` docstring for information about how `SED` normalization affects
        `ChromaticObject` normalization.

        Parameters:
            target_flux:    The desired flux normalization of the SED.
            bandpass:       A `Bandpass` object defining a filter bandpass.

        Returns:
            the new normalized `SED`.
        """
        current_flux = self.calculateFlux(bandpass)
        norm = target_flux/current_flux
        return self * norm

    def withMagnitude(self, target_magnitude, bandpass):
        """Return a new `SED` with magnitude through the `Bandpass` ``bandpass`` set to
        ``target_magnitude``.

        Note that this requires ``bandpass`` to have been assigned a zeropoint using
        `Bandpass.withZeropoint`.  See `ChromaticObject` docstring for information about how `SED`
        normalization affects `ChromaticObject` normalization.

        Parameters:
            target_magnitude:   The desired magnitude of the `SED`.
            bandpass:           A `Bandpass` object defining a filter bandpass.

        Returns:
            the new normalized `SED`.
        """
        if bandpass.zeropoint is None:
            raise GalSimError("Cannot call SED.withMagnitude on this bandpass, because it does "
                              "not have a zeropoint.  See Bandpass.withZeropoint()")
        current_magnitude = self.calculateMagnitude(bandpass)
        norm = 10**(-0.4*(target_magnitude - current_magnitude))
        return self * norm

    def atRedshift(self, redshift):
        """Return a new `SED` with redshifted wavelengths.

        Parameters:
            redshift:   The redshift for the returned `SED`

        Returns:
            the redshifted `SED`.
        """
        if redshift == self.redshift:
            return self
        if redshift <= -1:
            raise GalSimRangeError("Invalid redshift", redshift, -1.)
        zfactor = (1.0 + redshift) / (1.0 + self.redshift)
        wave_list = self.wave_list * zfactor
        blue_limit = self.blue_limit * zfactor
        red_limit = self.red_limit * zfactor

        return SED(self._spec, self.wave_type, self.flux_type, redshift, self.fast,
                   _wave_list=wave_list, _blue_limit=blue_limit, _red_limit=red_limit)

    def calculateFlux(self, bandpass):
        """Return the flux (photons/cm^2/s) of the `SED` through the `Bandpass` bandpass.

        Parameters:
            bandpass:   A `Bandpass` object representing a filter, or None to compute the
                        bolometric flux.  For the bolometric flux the integration limits will be
                        set to (0, infinity), which implies that the `SED` needs to be evaluable
                        over this entire range.

        Returns:
            the flux through the bandpass.
        """
        if self.dimensionless:
            raise GalSimSEDError("Cannot calculate flux of dimensionless SED.", self)
        if bandpass is None: # compute bolometric flux
            bandpass = Bandpass(lambda w: 1., 'nm', blue_limit=0., red_limit=SED._bolo_max_wave)
        if len(bandpass.wave_list) > 0 or len(self.wave_list) > 0:
            slop = 1e-6 # nm
            if (self.blue_limit > bandpass.blue_limit + slop
                    or self.red_limit < bandpass.red_limit - slop):
                raise GalSimRangeError("Bandpass is not completely within defined wavelength "
                                       "range for this SED.",
                                       (bandpass.blue_limit, bandpass.red_limit),
                                       self.blue_limit, self.red_limit)
            wmin = max(self.blue_limit, bandpass.blue_limit)
            wmax = min(self.red_limit, bandpass.red_limit)
            if self.fast and isinstance(self._fast_spec, LookupTable):
                wf = 1./(1.+self.redshift) / bandpass.wave_factor
                ff = 1./bandpass.wave_factor
                wmin *= bandpass.wave_factor
                wmax *= bandpass.wave_factor
                return self._fast_spec.integrate_product(bandpass._tp, wmin, wmax, wf) * ff
            else:
                w, _, _ = utilities.combine_wave_list(self, bandpass)
                if not self.fast and self.flux_type != 'fphotons':
                    # When not fast, the SED definition is not linear between the wave_list
                    # points, so this can be slightly inaccurate if the waves are too far apart.
                    # Add in 100 uniformly spaced points to achieve relative accurace ~few e-6.
                    w = utilities.merge_sorted([w, np.linspace(w[0], w[-1], 100)])
                interpolant = (bandpass._tp.interpolant if hasattr(bandpass._tp, 'interpolant')
                                else 'linear')
                return _LookupTable(w, bandpass(w), interpolant).integrate_product(self)
        else:
            return integ.int1d(lambda w: bandpass(w)*self(w),
                               bandpass.blue_limit, bandpass.red_limit)

    def calculateMagnitude(self, bandpass):
        """Return the `SED` magnitude through a `Bandpass` ``bandpass``.

        Note that this requires ``bandpass`` to have been assigned a zeropoint using
        `Bandpass.withZeropoint`.

        Parameters:
            bandpass:     A `Bandpass` object representing a filter, or None to compute the
                          bolometric magnitude.  For the bolometric magnitude the integration
                          limits will be set to (0, infinity), which implies that the `SED` needs
                          to be evaluable over this entire range.

        Returns:
            the bandpass magnitude.
        """
        if self.dimensionless:
            raise GalSimSEDError("Cannot calculate magnitude of dimensionless SED.", self)
        if bandpass.zeropoint is None:
            raise GalSimError("Cannot do this calculation for a bandpass without an assigned "
                              "zeropoint")
        flux = self.calculateFlux(bandpass)
        return -2.5 * np.log10(flux) + bandpass.zeropoint

    def thin(self, rel_err=1.e-4, trim_zeros=True, preserve_range=True, fast_search=True):
        """Remove some tabulated values while keeping the integral over the set of tabulated values
        still accurate to ``rel_err``.

        This is only relevant if the `SED` was initialized with a `LookupTable` or from a file
        (which internally creates a `LookupTable`).

        Parameters:
            rel_err:          The relative error allowed in the integral over the `SED`
                              [default: 1.e-4]
            trim_zeros:       Remove redundant leading and trailing points where f=0?  (The last
                              leading point with f=0 and the first trailing point with f=0 will
                              be retained).  Note that if both trim_leading_zeros and
                              preserve_range are True, then the only the range of ``x`` *after*
                              zero trimming is preserved.  [default: True]
            preserve_range:   Should the original range (``blue_limit`` and ``red_limit``) of the
                              `SED` be preserved? (True) Or should the ends be trimmed to
                              include only the region where the integral is significant? (False)
                              [default: True]
            fast_search:      If set to True, then the underlying algorithm will use a
                              relatively fast O(N) algorithm to select points to include in the
                              thinned approximation.  If set to False, then a slower O(N^2)
                              algorithm will be used.  We have found that the slower algorithm
                              tends to yield a thinned representation that retains fewer samples
                              while still meeting the relative error requirement.
                              [default: True]

        Returns:
            the thinned `SED`.
        """
        if len(self.wave_list) > 0:
            rest_wave_native = self._get_rest_native_waves(self.wave_list)
            spec_native = self._spec(rest_wave_native)

            # Note that this is thinning in native units, not nm and photons/nm.
            interpolant = (self.interpolant if not isinstance(self._spec, LookupTable)
                           else self._spec.interpolant)
            newx, newf = utilities.thin_tabulated_values(
                    rest_wave_native, spec_native, rel_err=rel_err,
                    trim_zeros=trim_zeros, preserve_range=preserve_range,
                    fast_search=fast_search, interpolant=interpolant)

            newspec = _LookupTable(newx, newf, interpolant=interpolant)
            return SED(newspec, self.wave_type, self.flux_type, redshift=self.redshift,
                       fast=self.fast)
        else:
            return self

    def calculateDCRMomentShifts(self, bandpass, **kwargs):
        """Calculates shifts in first and second moments of PSF due to differential chromatic
        refraction (DCR).

        I.e., equations (1) and (2) from Plazas and Bernstein (2012):

        http://arxiv.org/abs/1204.1346).

        Parameters:
            bandpass:           `Bandpass` through which object is being imaged.
            zenith_angle:       `Angle` from object to zenith
            parallactic_angle:  Parallactic angle, i.e. the position angle of the zenith,
                                measured from North through East.  [default: 0]
            obj_coord:          Celestial coordinates of the object being drawn as a
                                `CelestialCoord`. [default: None]
            zenith_coord:       Celestial coordinates of the zenith as a `CelestialCoord`.
                                [default: None]
            HA:                 Hour angle of the object as an `Angle`. [default: None]
            latitude:           Latitude of the observer as an `Angle`. [default: None]
            pressure:           Air pressure in kiloPascals.  [default: 69.328 kPa]
            temperature:        Temperature in Kelvins.  [default: 293.15 K]
            H2O_pressure:       Water vapor pressure in kiloPascals.  [default: 1.067 kPa]

        Returns:
            a tuple:

            - The first element is the vector of DCR first moment shifts
            - The second element is the 2x2 matrix of DCR second (central) moment shifts.
        """
        if self.dimensionless:
            raise GalSimSEDError("Cannot calculate DCR shifts of dimensionless SED.", self)

        zenith_angle, parallactic_angle, kwargs = dcr.parse_dcr_angles(**kwargs)

        # Any remaining kwargs will get forwarded to galsim.dcr.get_refraction
        # Check that they're valid
        for kw in kwargs:
            if kw not in ('temperature', 'pressure', 'H2O_pressure'):
                raise (TypeError("Got unexpected keyword in calculateDCRMomentShifts: {0}"
                                 .format(kw)))

        # Now actually start calculating things.
        flux = self.calculateFlux(bandpass)
        if len(self.wave_list) > 0 or len(bandpass.wave_list) > 0:
            w, _, _ = utilities.combine_wave_list(self, bandpass)
            interpolant = (bandpass._tp.interpolant if hasattr(bandpass._tp, 'interpolant')
                            else 'linear')
            bp = _LookupTable(w, bandpass(w), interpolant)
            R = lambda w: dcr.get_refraction(w, zenith_angle, **kwargs)
            Rbar = bp.integrate_product(lambda w: self(w) * R(w)) / flux
            V = bp.integrate_product(lambda w: self(w) * (R(w)-Rbar)**2) / flux
        else:
            weight = lambda w: bandpass(w) * self(w)
            Rbar_kernel = lambda w: dcr.get_refraction(w, zenith_angle, **kwargs)
            Rbar = integ.int1d(lambda w: weight(w) * Rbar_kernel(w),
                               bandpass.blue_limit, bandpass.red_limit)
            Rbar /= flux
            V_kernel = lambda w: (dcr.get_refraction(w, zenith_angle, **kwargs) - Rbar)**2
            V = integ.int1d(lambda w: weight(w) * V_kernel(w),
                            bandpass.blue_limit, bandpass.red_limit)
            V /= flux
        # Rbar and V are computed above assuming that the parallactic angle is 0.  Hence we
        # need to rotate our frame by the parallactic angle to get the desired output.
        sinp, cosp = parallactic_angle.sincos()
        rot = np.array([[cosp, -sinp], [sinp, cosp]])
        Rbar = Rbar * rot.dot(np.array([0,1]))
        V = rot.dot(np.array([[0, 0], [0, V]])).dot(rot.T)
        return Rbar, V

    def calculateSeeingMomentRatio(self, bandpass, alpha=-0.2, base_wavelength=500):
        """Calculates the relative size of a PSF compared to the monochromatic PSF size at
        wavelength ``base_wavelength``.

        Parameters:
            bandpass:           `Bandpass` through which object is being imaged.
            alpha:              Power law index for wavelength-dependent seeing.  [default:
                                -0.2, the prediction for Kolmogorov turbulence]
            base_wavelength:    Reference wavelength in nm from which to compute the relative
                                PSF size.  [default: 500]

        Returns:
            the ratio of the PSF second moments to the second moments of the reference PSF.
        """
        if self.dimensionless:
            raise GalSimSEDError("Cannot calculate seeing moment ratio of dimensionless SED.", self)
        flux = self.calculateFlux(bandpass)
        if len(self.wave_list) > 0 or len(bandpass.wave_list) > 0:
            # With three things multiplied together, we can't rely on integrate_product
            # being completely accurate if the waves are spaced too far apart, especially with
            # a power law being one of the factors.
            # So make sure to include a uniform density of points along with the native sed and
            # bandpass points. The error goes like dx**3, so 100 points should give relative
            # errors of order ~few e-6.
            w, _, _ = utilities.combine_wave_list([self, bandpass])
            w = utilities.merge_sorted([w, np.linspace(w[0], w[-1], 100)])
            interpolant = (bandpass._tp.interpolant if hasattr(bandpass._tp, 'interpolant')
                            else 'linear')
            bp = _LookupTable(w, bandpass(w), interpolant)
            return bp.integrate_product(lambda w: self(w) * (w/base_wavelength)**(2*alpha)) / flux
        else:
            weight = lambda w: bandpass(w) * self(w)
            kernel = lambda w: (w/base_wavelength)**(2*alpha)
            return integ.int1d(lambda w: weight(w) * kernel(w),
                               bandpass.blue_limit, bandpass.red_limit) / flux

    @lazy_property
    def _cache_deviate(self):
        return dict()

    def sampleWavelength(self, nphotons, bandpass, rng=None, npoints=None):
        """Sample a number of random wavelength values from the `SED`, possibly as observed through
        a `Bandpass` bandpass.

        Parameters:
            nphotons:    Number of samples (photons) to randomly draw.
            bandpass:    A `Bandpass` object representing a filter, or None to sample over the full
                         `SED` wavelength range.
            rng:         If provided, a random number generator that is any kind of `BaseDeviate`
                         object. If ``rng`` is None, one will be automatically created from the
                         system. [default: None]
            npoints:     Number of points `DistDeviate` should use for its internal interpolation
                         tables. [default: None, which uses the `DistDeviate` default]
        """
        nphotons=int(nphotons)

        key = (bandpass,npoints)
        if key in self._cache_deviate:
            dev = self._cache_deviate[key]
        else:
            if bandpass is None:
                sed = self
            else:
                sed = self._mul_bandpass(bandpass)

            xmin = sed.blue_limit / (1.+self.redshift)
            xmax = sed.red_limit / (1.+self.redshift)
            dev = DistDeviate(function=sed._fast_spec, x_min=xmin, x_max=xmax,
                              npoints=npoints, clip_neg=True)
            self._cache_deviate[key] = dev

        # Reset the deviate explicitly
        if rng is not None: dev.reset(rng)

        ret = np.empty(nphotons)
        dev.generate(ret)

        if self.redshift != 0:
            ret *= (1. + self.redshift)
            # Rarely, with the redshift round trip, this can produce wavelengths < blue_limit.
            # If this happens, set those values equal to blue_limit.
            # I'm not sure if the red limit overrun can happen (we didn't see any in the use case
            # that noticed the blue overruns), but it seems prudent to also correct any of these
            # that may occur too.  Plus it's not noticeably slower using clip to do both at once.
            if bandpass is not None:
                np.clip(ret, bandpass.blue_limit, bandpass.red_limit, out=ret)

        return ret

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, SED) and
                 self._orig_spec == other._orig_spec and
                 self.fast == other.fast and
                 self.wave_type == other.wave_type and
                 self.flux_type == other.flux_type and
                 self.redshift == other.redshift and
                 self.red_limit == other.red_limit and
                 self.blue_limit == other.blue_limit and
                 np.array_equal(self.wave_list,other.wave_list)))

    def __ne__(self, other): return not self.__eq__(other)

    def __hash__(self):
        # Cache this in case self._orig_spec or self.wave_list is long.
        if not hasattr(self, '_hash'):
            self._hash = hash(("galsim.SED", self._orig_spec, self.wave_type, self.flux_type,
                               self.redshift, self.fast, self.blue_limit, self.red_limit,
                               tuple(self.wave_list)))
        return self._hash

    def __repr__(self):
        outstr = ('galsim.SED(%r, wave_type=%r, flux_type=%r, redshift=%r, fast=%r, '
                  'interpolant=%r, _wave_list=%r, _blue_limit=%r, _red_limit=%s)')%(
                      self._orig_spec, self.wave_type, self._flux_type, self.redshift, self.fast,
                      self.interpolant, self.wave_list, self.blue_limit,
                      "float('inf')" if self.red_limit == np.inf else repr(self.red_limit))
        return outstr

    def __str__(self):
        orig_spec = repr(self._orig_spec)
        if len(orig_spec) > 80:
            orig_spec = str(self._orig_spec)
        return 'galsim.SED(%s, redshift=%s)'%(orig_spec, self.redshift)

    def __getstate__(self):
        d = self.__dict__.copy()
        if not isinstance(d['_spec'], LookupTable):
            del d['_spec']
        d.pop('_fast_spec',None)
        del d['_call']
        del d['_get_native_waves']
        del d['_get_rest_native_waves']
        del d['_flux_to_photons']
        return d

    def __setstate__(self, d):
        self.__dict__ = d
        if '_spec' not in d:
            self._initialize_spec()
        self._setup_funcs()


class EmissionLine(SED):
    """Emission line SED.

    Creates a simple triangle-shaped emission line with a given wavelength and
    FWHM.

    Parameters:
        wavelength:    The wavelength of the line.
        fwhm:          The full-width-half-max of the line.  [default: 1.0]
        flux:          The integrated flux of the line.  [default: 1.0]
        wave_type:     The units of ``wavelength`` and ``fwhm`` above.  See SED
                       constructor for options.  [default: 'nm']
        flux_type:     The units of flux used for this SED.  See SED constructor
                       for options.  [default: 'fphotons']
        max_wave       The maximum wavelength to use in the LookupTable for the SED.
                       [default: {}; must be > wavelength+fwhm]
        **kwargs:      Any additional keyword arguments to pass to the `SED`
                       constructor.
    """.format(SED._bolo_max_wave)
    def __init__(
        self,
        wavelength,
        fwhm=1.0,
        flux=1.0,
        wave_type='nm',
        flux_type='fphotons',
        max_wave=SED._bolo_max_wave,
        **kwargs
    ):
        self.wavelength = wavelength
        self.fwhm = fwhm
        self.flux = flux
        _, wave_factor = SED._parse_wave_type(wave_type)
        if wave_factor is None:
            raise GalSimValueError("wave_type must be a distance unit", wave_type)
        assert max_wave > wavelength + fwhm

        w = wavelength
        # Some operations can turn a 0 into 1.e-15, which can lead to large errors
        # when integrating from w+fwhm to max_wave.  So add a second set of 0's at
        # w +- 2*fwhm to make sure it's exactly 0 for most of the range.
        spec = LookupTable(
            [0.0, w-2*fwhm, w-fwhm, w, w+fwhm, w+2*fwhm, max_wave*wave_factor],
            [0, 0, 0, flux/fwhm, 0, 0, 0],
            interpolant='linear'
        )
        super().__init__(
            spec,
            wave_type=wave_type,
            flux_type=flux_type,
            **kwargs
        )

    def atRedshift(self, redshift):
        """Return a new `EmissionLine` with redshifted wavelengths.

        Parameters:
            redshift:   The redshift for the returned `EmissionLine`

        Returns:
            the redshifted `EmissionLine`.
        """
        if redshift == self.redshift:
            return self
        if redshift <= -1:
            raise GalSimRangeError("Invalid redshift", redshift, -1.)
        return EmissionLine(
            self.wavelength,
            self.fwhm,
            flux=self.flux,
            wave_type=self.wave_type,
            flux_type=self.flux_type,
            redshift=redshift,
            fast=self.fast
        )

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            flux = self.flux * other
            return EmissionLine(
                self.wavelength,
                self.fwhm,
                flux=flux,
                wave_type=self.wave_type,
                flux_type=self.flux_type,
                redshift=self.redshift,
                fast=self.fast
            )
        else:
            return super().__mul__(other)

    def __div__(self, other):
        if isinstance(other, (int, float)):
            flux = self.flux / other
            return EmissionLine(
                self.wavelength,
                self.fwhm,
                flux=flux,
                wave_type=self.wave_type,
                flux_type=self.flux_type,
                redshift=self.redshift,
                fast=self.fast
            )
        else:
            return super().__div__(other)

    __truediv__ = __div__

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, EmissionLine) and
                 self.wavelength == other.wavelength and
                 self.fwhm == other.fwhm and
                 self.flux == other.flux and
                 self.wave_type == other.wave_type and
                 self.flux_type == other.flux_type and
                 self.redshift == other.redshift))

    def __hash__(self):
        return hash(("galsim.EmissionLine", self.wavelength, self.fwhm, self.flux))

    def __repr__(self):
        outstr = ('galsim.EmissionLine(%r, %r, flux=%r, wave_type=%r, flux_type=%r, redshift=%r,'
                  ' fast=%r)')%(
                      self.wavelength, self.fwhm, self.flux, self.wave_type, self._flux_type,
                      self.redshift, self.fast)
        return outstr

    def __str__(self):
        outstr = 'galsim.EmissionLine(wavelength=%s, fwhm=%s'%(self.wavelength, self.fwhm)
        if self.flux != 1.0:
            outstr += ', flux=%s'%self.flux
        if self.redshift != 0.0:
            outstr += ', redshift=%s'%self.redshift
        outstr += ')'
        return outstr


# Put this at the bottom to avoid circular import errors.
from .bandpass import Bandpass



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import logging
import numpy as np
import math

from .util import LoggerWrapper, GetIndex, GetRNG
from .value import ParseValue, GetCurrentValue, GetAllParams
from .input import RegisterInputConnectedType
from ..errors import GalSimConfigError, GalSimConfigValueError
from ..image import Image
from ..position import PositionI
from ..random import PoissonDeviate
from ..noise import GaussianNoise, PoissonNoise, DeviateNoise, CCDNoise
from ..correlatednoise import getCOSMOSNoise, BaseCorrelatedNoise, UncorrelatedNoise

# This file handles the functionality for adding noise and the sky to an image after
# drawing the objects.

# This module-level dict will store all the registered noise types.
# See the RegisterNoiseType function at the end of this file.
# The keys are the (string) names of the noise types, and the values will be builder objects
# that will perform the different functions related to adding noise to images.
valid_noise_types = {}

#
# First the driver functions:
#

def AddSky(config, im):
    """Add the sky level to the image

    Parameters:
        config:         The (base) configuration dict
        im:             The image onto which to add the sky
    """
    index, orig_index_key = GetIndex(config['image'], config)
    config['index_key'] = 'image_num'
    sky = GetSky(config['image'], config, full=True)
    if sky:
        im += sky
    config['index_key'] = orig_index_key


def AddNoise(config, im, current_var=0., logger=None):
    """
    Add noise to an image according to the noise specifications in the noise dict.

    Parameters:
        config:         The (base) configuration dict
        im:             The image onto which to add the noise
        current_var:    The current noise variance present in the image already [default: 0]
        logger:         If given, a logger object to log progress. [default: None]

    Returns:
        Variance added to the image (units are ADU if gain != 1)
    """
    from .stamp import SetupConfigObjNum

    logger = LoggerWrapper(logger)
    if 'noise' in config['image']:
        noise = config['image']['noise']
    else: # No noise.
        return 0
    if not isinstance(noise, dict):
        raise GalSimConfigError("image.noise is not a dict.")

    # Default is Poisson
    noise_type = noise.get('type', 'Poisson')
    if noise_type not in valid_noise_types:
        raise GalSimConfigValueError("Invalid noise.type.", noise_type,
                                     list(valid_noise_types.keys()))

    # We need to use image_num for the index_key, but if we are running this from the stamp
    # building phase, then we want to use obj_num_rng for the noise rng.  So get the rng now
    # before we change config['index_key'].
    index, orig_index_key = GetIndex(noise, config)
    rng = GetRNG(noise, config)

    # This makes sure draw_method is properly copied over and given a default value.
    SetupConfigObjNum(config, config.get('obj_num',0), logger)
    draw_method = GetCurrentValue('draw_method', config['stamp'], str, config)

    builder = valid_noise_types[noise_type]
    config['index_key'] = 'image_num'
    var = builder.addNoise(noise, config, im, rng, current_var, draw_method, logger)
    config['index_key'] = orig_index_key

    return var

def CalculateNoiseVariance(config, full=False):
    """
    Calculate the noise variance from the noise specified in the noise dict.

    Parameters:
        config:     The (base) configuration dict
        full:       If the noise is variable across the image, return the full image with the
                    noise variance at every pixel.  Otherwise, just return the value at the center.

    Returns:
        the noise variance (units are ADU if gain != 1)
    """
    if 'noise' in config['image']:
        noise = config['image']['noise']
    else: # No noise.
        return 0
    if not isinstance(noise, dict):
        raise GalSimConfigError("image.noise is not a dict.")

    noise_type = noise.get('type', 'Poisson')
    if noise_type not in valid_noise_types:
        raise GalSimConfigValueError("Invalid noise.type.", noise_type,
                                     list(valid_noise_types.keys()))

    index, orig_index_key = GetIndex(noise, config)
    config['index_key'] = 'image_num'

    builder = valid_noise_types[noise_type]
    var = builder.getNoiseVariance(noise, config, full=full)
    config['index_key'] = orig_index_key

    return var

def AddNoiseVariance(config, im, include_obj_var=False, logger=None):
    """
    Add the noise variance to an image according to the noise specifications in the noise dict.
    Typically, this is used for building a weight map, which is typically the inverse variance.

    Parameters:
        config:             The (base) configuration dict
        im:                 The image onto which to add the variance values
        include_obj_var:    Whether to add the variance from the object photons for noise
                            models that have a component based on the number of photons.
                            Note: if this is True, the returned variance will not include this
                            contribution to the noise variance.  [default: False]
        logger:             If given, a logger object to log progress. [default: None]

    Returns:
        the variance in the image (units are ADU if gain != 1)
    """
    logger = LoggerWrapper(logger)
    if 'noise' in config['image']:
        noise = config['image']['noise']
    else: # No noise.
        return
    if not isinstance(noise, dict):
        raise GalSimConfigError("image.noise is not a dict.")

    noise_type = noise.get('type', 'Poisson')
    if noise_type not in valid_noise_types:
        raise GalSimConfigValueError("Invalid noise.type.", noise_type,
                                     list(valid_noise_types.keys()))

    index, orig_index_key = GetIndex(noise, config)
    config['index_key'] = 'image_num'

    builder = valid_noise_types[noise_type]
    builder.addNoiseVariance(noise, config, im, include_obj_var, logger)
    config['index_key'] = orig_index_key

def GetSky(config, base, logger=None, full=False):
    """Parse the sky information and return either a float value for the sky level per pixel
    or an image, as needed.

    If an image is required (because wcs is not uniform) then it will use the presence of
    base['image_pos'] to determine what size image to return (stamp or full).  If there is
    a current image_pos, then we are doing a stamp.  Otherwise a full image.

    Parameters:
        config:             The configuration field with the sky specification, which can be either
                            base['image'] or base['image']['noise']
        base:               The base configuration dict
        logger:             If given, a logger object to log progress. [default: None]
        full:               If the sky level is variable across the image, return the full
                            image with the sky at every pixel.  Otherwise, just return the
                            sky at the image center.

    Returns:
        sky, either a float value or an Image.  (The latter only if full=True)
    """
    logger = LoggerWrapper(logger)
    if 'sky_level' in config:
        if 'sky_level_pixel' in config:
            raise GalSimConfigValueError(
                "Cannot specify both sky_level and sky_level_pixel",
                (config['sky_level'], config['sky_level_pixel']))
        sky_level = ParseValue(config,'sky_level',base,float)[0]
        logger.debug('image %d, obj %d: sky_level = %f',
                     base.get('image_num',0),base.get('obj_num',0), sky_level)
        wcs = base['wcs']
        if wcs._isUniform:
            sky_level_pixel = sky_level * wcs.pixelArea()
            logger.debug('image %d, obj %d: Uniform: sky_level_pixel = %f',
                         base.get('image_num',0),base.get('obj_num',0), sky_level_pixel)
            return sky_level_pixel
        elif full:
            # This calculation is non-trivial, so store this in case we need it again.
            tag = (id(base), base.get('file_num',0), base.get('image_num',0), base.get('obj_num',0))
            if config.get('_current_sky_tag',None) == tag:
                logger.debug('image %d, obj %d: Using saved sky image',
                             base.get('image_num',0),base.get('obj_num',0))
                return config['_current_sky']
            else:
                logger.debug('image %d, obj %d: Non-uniform wcs.  Making sky image',
                             base.get('image_num',0),base.get('obj_num',0))
                bounds = base['current_noise_image'].bounds
                sky = Image(bounds, wcs=wcs)
                wcs.makeSkyImage(sky, sky_level)
                sensor = base.get('sensor', None)
                if sensor is not None:
                    center = base.get('image_origin', PositionI(1,1)) - sky.origin
                    use_flux = config.get('use_flux_sky_areas', False)
                    # TODO: If use_flux_sky_areas = True, then we should really build this up
                    #       in steps.  E.g. for a flat field.
                    #       This one step calcualtion isn't right.
                    area = sensor.calculate_pixel_areas(sky, orig_center=center, use_flux=use_flux)
                    sky *= area
                config['_current_sky_tag'] = tag
                config['_current_sky'] = sky
                return sky
        else:
            center = base['current_noise_image'].bounds.true_center
            return wcs.local(image_pos=center).pixelArea() * sky_level
    elif 'sky_level_pixel' in config:
        sky_level_pixel = ParseValue(config,'sky_level_pixel',base,float)[0]
        logger.debug('image %d, obj %d: sky_level_pixel = %f',
                     base.get('image_num',0),base.get('obj_num',0), sky_level_pixel)
        return sky_level_pixel
    else:
        return 0.


# items that are parsed separately from the normal noise function
noise_ignore = [ 'whiten', 'symmetrize', 'use_flux_sky_areas' ]

class NoiseBuilder:
    """A base class for building noise objects and applying the noise to images.

    The base class doesn't do anything, but it defines the call signatures of the methods
    that derived classes should use for the different specific noise types.
    """
    def addNoise(self, config, base, im, rng, current_var, draw_method, logger):
        """Read the noise parameters from the config dict and add the appropriate noise to the
        given image.

        Parameters:
            config:         The configuration dict for the noise field.
            base:           The base configuration dict.
            im:             The image onto which to add the noise
            rng:            The random number generator to use for adding the noise.
            current_var:    The current noise variance present in the image already.
            draw_method:    The method that was used to draw the objects on the image.
            logger:         If given, a logger object to log progress.

        Returns:
            the variance of the noise model (units are ADU if gain != 1)
        """
        raise NotImplementedError("The %s class has not overridden addNoise"%self.__class__)

    def getNoiseVariance(self, config, base, full=False):
        """Read the noise parameters from the config dict and return the variance.

        Parameters:
            config:     The configuration dict for the noise field.
            base:       The base configuration dict.
            full:       If the noise is variable across the image, return the full image with the
                        noise variance at every pixel.  Otherwise, just return the value at the
                        center.

        Returns:
            the variance of the noise model (units are ADU if gain != 1)
        """
        raise NotImplementedError("The %s class has not overridden addNoise"%self.__class__)

    def addNoiseVariance(self, config, base, im, include_obj_var, logger):
        """Read the noise parameters from the config dict and add the appropriate noise variance
        to the given image.

        This is used for constructing the weight map iamge.  It doesn't add a random value to
        each pixel.  Rather, it adds the variance of the noise that was used in the main image to
        each pixel in this image.

        This method has a default implemenation that is appropriate for noise models that have
        a constant noise variance.  It just gets the variance from getNoiseVariance and adds
        that constant value to every pixel.

        Parameters:
            config:             The configuration dict for the noise field.
            base:               The base configuration dict.
            im:                 The image onto which to add the noise variance
            include_obj_var:    Whether the noise variance values should the photon noise from
                                object flux in addition to the sky flux.  Only relevant for
                                noise models that are based on the image flux values such as
                                Poisson and CCDNoise.
            logger:             If given, a logger object to log progress.
        """
        im += self.getNoiseVariance(config, base, full=True)

#
# Gaussian
#

class GaussianNoiseBuilder(NoiseBuilder):

    def addNoise(self, config, base, im, rng, current_var, draw_method, logger):

        # Read the noise variance
        var = self.getNoiseVariance(config, base)
        ret = var  # save for the return value

        # If we already have some variance in the image (from whitening), then we subtract this much
        # from sigma**2.
        if current_var:
            logger.debug('image %d, obj %d: Target variance is %f, current variance is %f',
                        base.get('image_num',0),base.get('obj_num',0),var,current_var)
            if var < current_var:
                raise GalSimConfigError(
                    "Whitening already added more noise than the requested Gaussian noise.")
            var -= current_var

        # Now apply the noise.
        sigma = math.sqrt(var)
        im.addNoise(GaussianNoise(rng,sigma=sigma))

        logger.debug('image %d, obj %d: Added Gaussian noise with var = %f',
                     base.get('image_num',0),base.get('obj_num',0),var)

        return ret

    def getNoiseVariance(self, config, base, full=False):

        # The noise level can be specified either as a sigma or a variance.  Here we just calculate
        # the value of the variance from either one.
        single = [ { 'sigma' : float , 'variance' : float } ]
        params = GetAllParams(config, base, single=single, ignore=noise_ignore)[0]
        if 'sigma' in params:
            sigma = params['sigma']
            return sigma * sigma
        else:
            return params['variance']


#
# Poisson
#

class PoissonNoiseBuilder(NoiseBuilder):

    def addNoise(self, config, base, im, rng, current_var, draw_method, logger):

        # Get how much extra sky to assume from the image.noise attribute.
        sky = GetSky(base['image'], base, logger, full=True)
        extra_sky = GetSky(config, base, logger, full=True)
        total_sky = sky + extra_sky # for the return value
        if isinstance(total_sky, Image):
            var = np.mean(total_sky.array)
        else:
            var = total_sky
        # (This could be zero, in which case we only add poisson noise for the object photons)

        # If we already have some variance in the image (from whitening), then we subtract this
        # much off of the sky level.  It's not precisely accurate, since the existing variance is
        # Gaussian, rather than Poisson, but it's the best we can do.
        if current_var:
            logger.debug('image %d, obj %d: Target variance is %f, current variance is %f',
                         base.get('image_num',0),base.get('obj_num',0), var, current_var)
            if isinstance(total_sky, Image):
                test = np.any(total_sky.array < current_var)
            else:
                test = (total_sky < current_var)
            if test:
                raise GalSimConfigError(
                    "Whitening already added more noise than the requested Poisson noise.")
            total_sky -= current_var
            extra_sky -= current_var

        # At this point, there is a slight difference between fft and phot. For photon shooting,
        # the galaxy already has Poisson noise, so we want to make sure not to add that again!
        if draw_method == 'phot':
            # Only add in the noise from the sky.
            if isinstance(total_sky, Image):
                noise_im = total_sky.copy()
                noise_im.addNoise(PoissonNoise(rng))
                noise_im -= total_sky
                # total_sky should now have zero mean, but with the noise of the total sky level.
                im += noise_im
            else:
                im.addNoise(DeviateNoise(PoissonDeviate(rng, mean=total_sky)))
                # This deviate adds a noisy version of the sky, so need to subtract the mean
                # back off.
                im -= total_sky
        else:
            # Do the normal PoissonNoise calculation.
            if isinstance(total_sky, Image):
                im += extra_sky
                im.addNoise(PoissonNoise(rng))
                im -= extra_sky
            else:
                im.addNoise(PoissonNoise(rng, sky_level=extra_sky))

        logger.debug('image %d, obj %d: Added Poisson noise',
                     base.get('image_num',0),base.get('obj_num',0))
        return var

    def getNoiseVariance(self, config, base, full=False):
        # The noise variance is the net sky level per pixel
        sky = GetSky(base['image'], base, full=full)
        sky += GetSky(config, base, full=full)
        return sky

    def addNoiseVariance(self, config, base, im, include_obj_var, logger):
        if include_obj_var:
            # The current image at this point should be the noise-free, sky-free image,
            # which is the object variance in each pixel.
            im += base['current_noise_image']

            # Note: For the phot case, we don't actually have an exact value for the variance in
            # each pixel, but the drawn image before adding the Poisson noise is our best guess for
            # the variance from the object's flux, so if we want the object variance included, this
            # is still the best we can do.

        # Add the total sky level
        im += self.getNoiseVariance(config, base, full=True)


#
# CCD
#

class CCDNoiseBuilder(NoiseBuilder):

    def getCCDNoiseParams(self, config, base):
        opt = { 'gain' : float , 'read_noise' : float }
        ignore = ['sky_level', 'sky_level_pixel']
        params = GetAllParams(config, base, opt=opt, ignore=noise_ignore + ignore)[0]
        gain = params.get('gain',1.0)
        read_noise = params.get('read_noise',0.0)
        read_noise_var = read_noise**2

        return gain, read_noise, read_noise_var

    def addNoise(self, config, base, im, rng, current_var, draw_method, logger):

        # This process goes a lot like the Poisson routine.  There are just two differences.
        # First, the Poisson noise is in electrons, not ADU, and now we allow for a gain = e-/ADU,
        # so we need to account for that properly.  Second, we also allow for an additional Gaussian
        # read noise.
        gain, read_noise, read_noise_var = self.getCCDNoiseParams(config, base)

        # Get how much extra sky to assume from the image.noise attribute.
        sky = GetSky(base['image'], base, logger, full=True)
        extra_sky = GetSky(config, base, logger, full=True)
        total_sky = sky + extra_sky
        if isinstance(total_sky, Image):
            var_adu = np.mean(total_sky.array) / gain + read_noise_var / gain**2
        else:
            var_adu = total_sky / gain + read_noise_var / gain**2


        # If we already have some variance in the image (from whitening), then we try to subtract
        # it from the read noise if possible.  If not, we subtract the rest off of the sky level.
        # It's not precisely accurate, since the existing variance is Gaussian, rather than
        # Poisson, but it's the best we can do.
        if current_var:
            logger.debug('image %d, obj %d: Target variance is %f, current variance is %f',
                        base.get('image_num',0),base.get('obj_num',0), var_adu, current_var)
            read_noise_var_adu = read_noise_var / gain**2
            if isinstance(total_sky, Image):
                test = np.any(total_sky.array/gain + read_noise_var_adu < current_var)
            else:
                logger.debug('image %d, obj %d: Target variance is %f, current variance is %f',
                             base.get('image_num',0),base.get('obj_num',0),
                             var_adu, current_var)
                test = var_adu < current_var
            if test:
                raise GalSimConfigError(
                    "Whitening already added more noise than the requested CCD noise.")
            if read_noise_var_adu >= current_var:
                # First try to take away from the read_noise, since this one is actually Gaussian.
                read_noise_var -= current_var * gain**2
                read_noise = math.sqrt(read_noise_var)
            else:
                # Take read_noise down to zero, since already have at least that much already.
                current_var -= read_noise_var_adu
                read_noise = 0
                read_noise_var = 0
                # Take the rest away from the sky level
                total_sky -= current_var * gain
                extra_sky -= current_var * gain

        # At this point, there is a slight difference between fft and phot. For photon shooting,
        # the galaxy already has Poisson noise, so we want to make sure not to add that again!
        if draw_method == 'phot':
            # Add in the noise from the sky.
            if isinstance(total_sky, Image):
                noise_im = total_sky.copy()
                if gain != 1.0: noise_im *= gain
                noise_im.addNoise(PoissonNoise(rng))
                if gain != 1.0: noise_im /= gain
                noise_im -= total_sky
                # total_sky should now have zero mean, but with the noise of the total sky level.
                im += noise_im
            else:
                if gain != 1.0: im *= gain
                pd = PoissonDeviate(rng, mean=total_sky*gain)
                im.addNoise(DeviateNoise(pd))
                if gain != 1.0: im /= gain
                im -= total_sky
            # And add the read noise
            if read_noise != 0.:
                im.addNoise(GaussianNoise(rng, sigma=read_noise/gain))
        else:
            # Do the normal CCDNoise calculation.
            if isinstance(total_sky, Image):
                im += extra_sky
                im.addNoise(CCDNoise(rng, gain=gain, read_noise=read_noise))
                im -= extra_sky
            else:
                im.addNoise(CCDNoise(rng, gain=gain, read_noise=read_noise, sky_level=extra_sky))

        logger.debug('image %d, obj %d: Added CCD noise with gain = %f, read_noise = %f',
                     base.get('image_num',0),base.get('obj_num',0), gain, read_noise)

        return var_adu

    def getNoiseVariance(self, config, base, full=False):
        # The noise variance is sky / gain + (read_noise/gain)**2
        gain, read_noise, read_noise_var = self.getCCDNoiseParams(config, base)

        # Start with the background sky level for the image
        sky = GetSky(base['image'], base, full=full)
        sky += GetSky(config, base, full=full)

        # Account for the gain and read_noise
        read_noise_var_adu = read_noise_var / gain**2
        return sky / gain + read_noise_var_adu

    def addNoiseVariance(self, config, base, im, include_obj_var, logger):
        gain, read_noise, read_noise_var = self.getCCDNoiseParams(config, base)
        if include_obj_var:
            # The current image at this point should be the noise-free, sky-free image,
            # which is the object variance in each pixel.
            if gain != 1.0:
                im += base['current_noise_image']/gain
            else:
                im += base['current_noise_image']

        # Now add on the regular CCDNoise from the sky and read noise.
        im += self.getNoiseVariance(config, base, full=True)


#
# Correlated
#

class CorrelatedNoiseBuilder(NoiseBuilder):

    def getNoise(self, config, base, rng=None):
        # Save the constructed CorrelatedNoise object, since we might need it again.
        tag = (id(base), base['file_num'], base['image_num'])
        if base.get('_current_cn_tag',None) == tag:
            return base['_current_cn']
        else:
            cn = self._readNoiseFile(config, base, rng)
            base['_current_cn'] = cn
            base['_current_cn_tag'] = tag
            return cn

    def _readNoiseFile(self, config, base, rng=None):
        req = { 'file_name': str, 'pixel_scale': float }
        opt = { 'variance' : float }
        kwargs = GetAllParams(config, base, req=req, opt=opt, ignore=noise_ignore)[0]
        if rng is None:
            rng = GetRNG(config, base)
        return BaseCorrelatedNoise.from_file(rng=rng, **kwargs)

    def addNoise(self, config, base, im, rng, current_var, draw_method, logger):

        # Build the correlated noise
        cn = self.getNoise(config,base,rng)
        var = cn.getVariance()

        # Subtract off the current variance if any
        if current_var:
            logger.debug('image %d, obj %d: Target variance is %f, current variance is %f',
                         base.get('image_num',0),base.get('obj_num',0), var, current_var)
            if var < current_var:
                raise GalSimConfigError(
                    "Whitening already added more noise than the requested COSMOS noise.")
            cn -= UncorrelatedNoise(current_var, rng=rng, wcs=cn.wcs)

        # Add the noise to the image
        im.addNoise(cn)

        logger.debug('image %d, obj %d: Added COSMOS correlated noise with variance = %f',
                     base.get('image_num',0),base.get('obj_num',0), var)
        return var

    def getNoiseVariance(self, config, base, full=False):
        cn = self.getNoise(config,base)
        return cn.getVariance()

# Specialization for COSMOS noise in particular (the OG version of this noise type).
class COSMOSNoiseBuilder(CorrelatedNoiseBuilder):
     def _readNoiseFile(self, config, base, rng=None):
        opt = { 'file_name' : str, 'cosmos_scale' : float, 'variance' : float }
        kwargs = GetAllParams(config, base, opt=opt, ignore=noise_ignore)[0]
        if rng is None:
            rng = GetRNG(config, base)
        return getCOSMOSNoise(rng=rng, **kwargs)


def RegisterNoiseType(noise_type, builder, input_type=None):
    """Register a noise type for use by the config apparatus.

    Parameters:
        noise_type:     The name of the type in config['image']['noise']
        builder:        A builder object to use for building the noise.  It should be an
                        instance of a subclass of NoiseBuilder.
        input_type:     If the builder utilises an input object, give the key name of the
                        input type here.  (If it uses more than one, this may be a list.)
                        [default: None]
    """
    valid_noise_types[noise_type] = builder
    RegisterInputConnectedType(input_type, noise_type)


RegisterNoiseType('Gaussian', GaussianNoiseBuilder())
RegisterNoiseType('Poisson', PoissonNoiseBuilder())
RegisterNoiseType('CCD', CCDNoiseBuilder())
RegisterNoiseType('Correlated', CorrelatedNoiseBuilder())
RegisterNoiseType('COSMOS', COSMOSNoiseBuilder())



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

# This file is logically part of utilities.py, but to avoid circular imports, we
# put a few things in here so files can import _utilties without triggering an ImportError.

import os
import functools
import weakref

from .errors import GalSimError, GalSimValueError

# Python 2/3 compatible definition of basestring without past.builtins
# (Based on this SO answer: https://stackoverflow.com/a/33699705/1332281)
basestring = ("".__class__, u"".__class__, b"".__class__)

class lazy_property:
    """
    This decorator will act similarly to @property, but will be efficient for multiple access
    to values that require some significant calculation.

    It works by replacing the attribute with the computed value, so after the first access,
    the property (an attribute of the class) is superseded by the new attribute of the instance.

    Note that is should only be used for non-mutable data, since the calculation will not be
    repeated if anything about the instance changes.

    Usage::

        @lazy_property
        def slow_function_to_be_used_as_a_property(self):
            x =  ...  # Some slow calculation.
            return x

    Base on an answer from http://stackoverflow.com/a/6849299
    """
    def __init__(self, fget):
        self.fget = fget
        self.func_name = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = self.fget(obj)
        setattr(obj, self.func_name, value)
        return value

class doc_inherit:
    '''
    This decorator will grab a doc string from a base class version of a method.
    Useful if the subclass doesn't change anything about the method API, but just has
    a specialized implementation.  This lets the documentation live only in one place.

    Usage::

        class Base:
            def some_method(self):
                """A nice description of the functionality
                """
                pass

        class Sub(Base):

            @doc_inherit
            def some_method(self):
                # Don't bother with any doc string here.
                pass

    Based on the Docstring Inheritance Decorator at:

    https://github.com/ActiveState/code/wiki/Python_index_1

    Although I (MJ) modified it slightly, since the original recipe there had a bug that made it
    not work properly with 2 levels of sub-classing (e.g. Pixel <- Box <- GSObject).
    '''
    def __init__(self, mthd):
        self.mthd = mthd
        self.name = mthd.__name__

    def __get__(self, obj, cls):
        for parent in cls.__bases__: # pragma: no branch
            parfunc = getattr(parent, self.name, None)
            if parfunc and getattr(parfunc, '__doc__', None): # pragma: no branch
                break

        if obj:
            return self.get_with_inst(obj, cls, parfunc)
        else:
            return self.get_no_inst(cls, parfunc)

    def get_with_inst(self, obj, cls, parfunc):
        @functools.wraps(self.mthd, assigned=('__name__','__module__'))
        def f(*args, **kwargs):
            return self.mthd(obj, *args, **kwargs)
        return self.use_parent_doc(f, parfunc)

    def get_no_inst(self, cls, parfunc):
        @functools.wraps(self.mthd, assigned=('__name__','__module__'))
        def f(*args, **kwargs): # pragma: no cover (without inst, this is not normally called.)
            return self.mthd(*args, **kwargs)
        return self.use_parent_doc(f, parfunc)

    def use_parent_doc(self, func, source):
        if source is None: # pragma: no cover
            raise NameError("Can't find '%s' in parents"%self.name)
        func.__doc__ = source.__doc__
        return func


def isinteger(value):
    """Check if a value is an integer type (including np.int64, long, etc.)

    Specifically, it checks whether value == int(value).

    Parameter:
        value:      The value to be checked whether it is an integer

    Returns:
        True if the value is an integer type, False otherwise.
    """
    try:
        return value == int(value)
    except TypeError:
        return False

def ensure_dir(target):
    """
    Make sure the directory for the target location exists, watching for a race condition

    In particular check if the OS reported that the directory already exists when running
    makedirs, which can happen if another process creates it before this one can

    Parameter:
        target:     The file name for which to ensure that all necessary directories exist.
    """
    _ERR_FILE_EXISTS=17
    dir = os.path.dirname(target)
    if dir == '': return

    exists = os.path.exists(dir)
    if not exists:
        try:
            os.makedirs(dir)
        except OSError as err:  # pragma: no cover
            # check if the file now exists, which can happen if some other
            # process created the directory between the os.path.exists call
            # above and the time of the makedirs attempt.  This is OK
            if err.errno != _ERR_FILE_EXISTS:
                raise err

    elif exists and not os.path.isdir(dir):
        raise OSError("tried to make directory '%s' "
                      "but a non-directory file of that "
                      "name already exists" % dir)


class LRU_Cache:
    """Simplified Least Recently Used Cache.

    Mostly stolen from http://code.activestate.com/recipes/577970-simplified-lru-cache/,
    but added a method for dynamic resizing.  The least recently used cached item is
    overwritten on a cache miss.

    Parameters:
        user_function:  A python function to cache.
        maxsize:        Maximum number of inputs to cache.  [Default: 1024]

    Example::

        >>> def slow_function(*args) # A slow-to-evaluate python function
        >>>    ...
        >>>
        >>> v1 = slow_function(*k1)  # Calling function is slow
        >>> v1 = slow_function(*k1)  # Calling again with same args is still slow
        >>> cache = galsim.utilities.LRU_Cache(slow_function)
        >>> v1 = cache(*k1)  # Returns slow_function(*k1), slowly the first time
        >>> v1 = cache(*k1)  # Returns slow_function(*k1) again, but fast this time.
    """
    def __init__(self, user_function, maxsize=1024):
        # Link layout:     [PREV, NEXT, KEY, RESULT]
        self.root = [None, None, None, None]
        self.user_function = user_function
        self.cache = {}
        self.maxsize = maxsize
        self.clear()

    def clear(self):
        self.root = [None, None, None, None]
        root = self.root
        cache = self.cache
        cache.clear()
        maxsize = self.maxsize
        last = root
        for i in range(maxsize):
            key = object()
            cache[key] = last[1] = last = [last, root, key, None]
        root[0] = last

    def __call__(self, *key):
        cache = self.cache
        root = self.root
        link = cache.get(key)
        if link is not None:
            # Cache hit: move link to last position
            link_prev, link_next, _, result = link
            link_prev[1] = link_next
            link_next[0] = link_prev
            last = root[0]
            last[1] = root[0] = link
            link[0] = last
            link[1] = root
            return result
        # Cache miss: evaluate and insert new key/value at root, then increment root
        #             so that just-evaluated value is in last position.
        result = self.user_function(*key)
        root = self.root  # re-establish root in case user_function modified it due to recursion
        root[2] = key
        root[3] = result
        oldroot = root
        root = self.root = root[1]
        root[2], oldkey = None, root[2]
        root[3], oldvalue = None, root[3]
        del cache[oldkey]
        cache[key] = oldroot
        return result

    def resize(self, maxsize):
        """Resize the cache.

        Increasing the size of the cache is non-destructive, i.e., previously cached inputs remain
        in the cache.  Decreasing the size of the cache will necessarily remove items from the
        cache if the cache is already filled.  Items are removed in least recently used order.

        Parameters:
            maxsize:    The new maximum number of inputs to cache.
        """
        oldsize = self.maxsize
        if maxsize == oldsize:
            return
        else:
            root = self.root
            cache = self.cache
            if maxsize <= 0:
                raise GalSimValueError("Invalid maxsize", maxsize)
            if maxsize < oldsize:
                for i in range(oldsize - maxsize):
                    # Delete root.next
                    current_next_link = root[1]
                    new_next_link = root[1] = root[1][1]
                    new_next_link[0] = root
                    del cache[current_next_link[2]]
            else: #  maxsize > oldsize:
                for i in range(maxsize - oldsize):
                    # Insert between root and root.next
                    key = object()
                    cache[key] = link = [root, root[1], key, None]
                    root[1][0] = link
                    root[1] = link
        self.maxsize = maxsize


def math_eval(str, other_modules=()):
    """Evaluate a string that may include numpy, np, or math commands.

    Parameters:
        str:            The string to evaluate
        other_modules.  Other modules in addition to numpy, np, math to import as well.
                        Should be given as a list of strings.  [default: None]

    Returns:
        Whatever the string evaluates to.
    """
    gdict = globals().copy()
    exec('import galsim', gdict)
    exec('import numpy', gdict)
    exec('import numpy as np', gdict)

    exec('import math', gdict)
    exec('import coord', gdict)
    for m in other_modules:  # pragma: no cover  (We don't use this.)
        exec('import ' + m, gdict)

    # A few other things that show up in reprs, so useful to import here.
    exec('from numpy import array, uint16, uint32, int16, int32, float32, float64, complex64, complex128, ndarray',
         gdict)
    exec('from astropy.units import Unit', gdict)

    return eval(str, gdict)

class WeakMethod:
    """Wrap a method in a weakref.

    This is useful if you want to specialize a function if certain conditions hold.
    You can check those conditions and return one of several possible implementations as
    a `lazy_property`.

    Using just a normal ``weakref`` doesn't work, but this class will work.

    From http://code.activestate.com/recipes/81253-weakmethod/
    """
    def __init__(self, f):
        self.f = f.__func__
        self.c = weakref.ref(f.__self__)
    def __call__(self, *args):
        try:
            # If the reference is dead, self.c() will be None, so this will raise an
            # AttributeError: 'NoneType' object has no attribute ...
            # Hopefully the method itself won't raise an AttributeError for anything else.
            return self.f(self.c(), *args)
        except AttributeError:  # pragma: no cover
            raise GalSimError('Method called on dead object')
